# -*- coding: utf-8 -*-
import decimal
import os
import posixpath
import re
import struct
import warnings
from datetime import datetime
from enum import Enum
from xml.etree import ElementTree as etree

import exifread
import filetype

from lektor.reporter import reporter
from lektor.utils import get_dependent_url
from lektor.utils import locate_executable
from lektor.utils import portable_popen


# yay shitty library
datetime.strptime("", "")


class ThumbnailMode(Enum):
    FIT = "fit"
    CROP = "crop"
    STRETCH = "stretch"

    DEFAULT = "fit"

    @property
    def label(self):
        """The mode's label as used in templates."""
        warnings.warn(
            "ThumbnailMode.label is deprecated. (Use ThumbnailMode.value instead.)",
            DeprecationWarning,
        )
        return self.value

    @classmethod
    def from_label(cls, label):
        """Looks up the thumbnail mode by its textual representation."""
        warnings.warn(
            "ThumbnailMode.from_label is deprecated. "
            "Use the ThumbnailMode constructor, "
            'e.g. "ThumbnailMode(label)", instead.',
            DeprecationWarning,
        )
        return cls(label)


def _convert_gps(coords, hem):
    deg, min, sec = [float(x.num) / float(x.den) for x in coords]
    sign = -1 if hem in "SW" else 1
    return sign * (deg + min / 60.0 + sec / 3600.0)


def _combine_make(make, model):
    make = make or ""
    model = model or ""
    if make and model.startswith(make):
        return make
    return " ".join([make, model]).strip()


_parse_svg_units_re = re.compile(
    r"(?P<value>[+-]?(?:\d+)(?:\.\d+)?)\s*(?P<unit>\D+)?", flags=re.IGNORECASE
)


def _parse_svg_units_px(length):
    match = _parse_svg_units_re.match(length)
    if not match:
        return None
    match = match.groupdict()
    if match["unit"] and match["unit"] != "px":
        return None
    try:
        return float(match["value"])
    except ValueError:
        return None


class EXIFInfo:
    def __init__(self, d):
        self._mapping = d

    def __bool__(self):
        return bool(self._mapping)

    __nonzero__ = __bool__

    def to_dict(self):
        rv = {}
        for key, value in self.__class__.__dict__.items():
            if key[:1] != "_" and isinstance(value, property):
                rv[key] = getattr(self, key)
        return rv

    def _get_string(self, key):
        try:
            value = self._mapping[key].values
        except KeyError:
            return None
        if isinstance(value, str):
            return value
        return value.decode("utf-8", "replace")

    def _get_int(self, key):
        try:
            return self._mapping[key].values[0]
        except LookupError:
            return None

    def _get_float(self, key, precision=4):
        try:
            val = self._mapping[key].values[0]
            if isinstance(val, int):
                return float(val)
            return round(float(val.num) / float(val.den), precision)
        except LookupError:
            return None

    def _get_frac_string(self, key):
        try:
            val = self._mapping[key].values[0]
            return "%s/%s" % (val.num, val.den)
        except LookupError:
            return None

    @property
    def artist(self):
        return self._get_string("Image Artist")

    @property
    def copyright(self):
        return self._get_string("Image Copyright")

    @property
    def camera_make(self):
        return self._get_string("Image Make")

    @property
    def camera_model(self):
        return self._get_string("Image Model")

    @property
    def camera(self):
        return _combine_make(self.camera_make, self.camera_model)

    @property
    def lens_make(self):
        return self._get_string("EXIF LensMake")

    @property
    def lens_model(self):
        return self._get_string("EXIF LensModel")

    @property
    def lens(self):
        return _combine_make(self.lens_make, self.lens_model)

    @property
    def aperture(self):
        return self._get_float("EXIF ApertureValue")

    @property
    def f_num(self):
        return self._get_float("EXIF FNumber")

    @property
    def f(self):
        return "ƒ/%s" % self.f_num

    @property
    def exposure_time(self):
        return self._get_frac_string("EXIF ExposureTime")

    @property
    def shutter_speed(self):
        val = self._get_float("EXIF ShutterSpeedValue")
        if val is not None:
            return "1/%d" % round(
                1 / (2 ** -val)  # pylint: disable=invalid-unary-operand-type
            )
        return None

    @property
    def focal_length(self):
        val = self._get_float("EXIF FocalLength")
        if val is not None:
            return "%smm" % val
        return None

    @property
    def focal_length_35mm(self):
        val = self._get_float("EXIF FocalLengthIn35mmFilm")
        if val is not None:
            return "%dmm" % val
        return None

    @property
    def flash_info(self):
        try:
            value = self._mapping["EXIF Flash"].printable
        except KeyError:
            return None
        if isinstance(value, str):
            return value
        return value.decode("utf-8")

    @property
    def iso(self):
        val = self._get_int("EXIF ISOSpeedRatings")
        if val is not None:
            return val
        return None

    @property
    def created_at(self):
        date_tags = (
            "GPS GPSDate",
            "Image DateTimeOriginal",
            "EXIF DateTimeOriginal",
            "EXIF DateTimeDigitized",
            "Image DateTime",
        )
        for tag in date_tags:
            try:
                return datetime.strptime(
                    self._mapping[tag].printable, "%Y:%m:%d %H:%M:%S"
                )
            except (KeyError, ValueError):
                continue
        return None

    @property
    def longitude(self):
        try:
            return _convert_gps(
                self._mapping["GPS GPSLongitude"].values,
                self._mapping["GPS GPSLongitudeRef"].printable,
            )
        except KeyError:
            return None

    @property
    def latitude(self):
        try:
            return _convert_gps(
                self._mapping["GPS GPSLatitude"].values,
                self._mapping["GPS GPSLatitudeRef"].printable,
            )
        except KeyError:
            return None

    @property
    def altitude(self):
        val = self._get_float("GPS GPSAltitude")
        if val is not None:
            try:
                ref = self._mapping["GPS GPSAltitudeRef"].values[0]
            except LookupError:
                ref = 0
            if ref == 1:
                val *= -1
            return val
        return None

    @property
    def location(self):
        lat = self.latitude
        long = self.longitude
        if lat is not None and long is not None:
            return (lat, long)
        return None

    @property
    def documentname(self):
        return self._get_string("Image DocumentName")

    @property
    def description(self):
        return self._get_string("Image ImageDescription")

    @property
    def is_rotated(self):
        """Return if the image is rotated according to the Orientation header.

        The Orientation header in EXIF stores an integer value between
        1 and 8, where the values 5-8 represent "portrait" orientations
        (rotated 90deg left, right, and mirrored versions of those), i.e.,
        the image is rotated.
        """
        return self._get_int("Image Orientation") in {5, 6, 7, 8}


def get_suffix(width, height, mode, quality=None):
    suffix = "" if width is None else str(width)
    if height is not None:
        suffix += "x%s" % height
    if mode != ThumbnailMode.DEFAULT:
        suffix += "_%s" % mode.value
    if quality is not None:
        suffix += "_q%s" % quality
    return suffix


def get_svg_info(fp):
    _, svg = next(etree.iterparse(fp, ["start"]), (None, None))
    fp.seek(0)
    width, height = None, None
    if svg is not None and svg.tag == "{http://www.w3.org/2000/svg}svg":
        width = _parse_svg_units_px(svg.attrib.get("width", ""))
        height = _parse_svg_units_px(svg.attrib.get("height", ""))
    return "svg", width, height


# see http://www.w3.org/Graphics/JPEG/itu-t81.pdf
# Table B.1 – Marker code assignments (page 32/36)
_JPEG_SOF_MARKERS = (
    # non-differential, Hufmann-coding
    0xC0,
    0xC1,
    0xC2,
    0xC3,
    # differential, Hufmann-coding
    0xC5,
    0xC6,
    0xC7,
    # non-differential, arithmetic-coding
    0xC9,
    0xCA,
    0xCB,
    # differential, arithmetic-coding
    0xCD,
    0xCE,
    0xCF,
)


def get_image_info(fp):
    """Reads some image info from a file descriptor."""
    head = fp.read(32)
    fp.seek(0)
    if len(head) < 24:
        return "unknown", None, None

    magic_bytes = b"<?xml", b"<svg"
    if any(map(head.strip().startswith, magic_bytes)):
        return get_svg_info(fp)

    _type = filetype.image_match(bytearray(head))
    fmt = _type.mime.split("/")[1] if _type else None

    width = None
    height = None
    if fmt == "png":
        check = struct.unpack(">i", head[4:8])[0]
        if check == 0x0D0A1A0A:
            width, height = struct.unpack(">ii", head[16:24])
    elif fmt == "gif":
        width, height = struct.unpack("<HH", head[6:10])
    elif fmt == "jpeg":
        # specification available under
        # http://www.w3.org/Graphics/JPEG/itu-t81.pdf
        # Annex B (page 31/35)

        # we are looking for a SOF marker ("start of frame").
        # skip over the "start of image" marker
        # (filetype detection took care of that).
        fp.seek(2)

        while True:
            byte = fp.read(1)

            # "All markers are assigned two-byte codes: an X’FF’ byte
            # followed by a byte which is not equal to 0 or X’FF’."
            if not byte or ord(byte) != 0xFF:
                raise Exception("Malformed JPEG image.")

            # "Any marker may optionally be preceded by any number
            # of fill bytes, which are bytes assigned code X’FF’."
            while ord(byte) == 0xFF:
                byte = fp.read(1)

            if ord(byte) not in _JPEG_SOF_MARKERS:
                # header length parameter takes 2 bytes for all markers
                length = struct.unpack(">H", fp.read(2))[0]
                fp.seek(length - 2, 1)
                continue

            # else...
            # see Figure B.3 – Frame header syntax (page 35/39) and
            # Table B.2 – Frame header parameter sizes and values
            # (page 36/40)
            fp.seek(3, 1)  # skip header length and precision parameters
            height, width = struct.unpack(">HH", fp.read(4))

            if height == 0:
                # "Value 0 indicates that the number of lines shall be
                # defined by the DNL marker [...]"
                #
                # DNL is not supported by most applications,
                # so we won't support it either.
                raise Exception("JPEG with DNL not supported.")

            break

        # if the file is rotated, we want, for all intents and purposes,
        # to return the dimensions swapped. (all client apps will display
        # the image rotated, and any template computations are likely to want
        # to make decisions based on the "visual", not the "real" dimensions.
        # thumbnail code also depends on this behaviour.)
        fp.seek(0)
        if is_rotated(fp):
            width, height = height, width
    else:
        fmt = None

    return fmt, width, height


def read_exif(fp):
    """Reads exif data from a file pointer of an image and returns it."""
    exif = exifread.process_file(fp)
    return EXIFInfo(exif)


def is_rotated(fp):
    """Fast version of read_exif(fp).is_rotated, using an exif header subset."""
    exif = exifread.process_file(fp, stop_tag="Orientation", details=False)
    return EXIFInfo(exif).is_rotated


def find_imagemagick(im=None):
    """Finds imagemagick and returns the path to it."""
    # If it's provided explicitly and it's valid, we go with that one.
    if im is not None and os.path.isfile(im):
        return im

    # On windows, imagemagick was renamed to magick, because
    # convert is system utility for fs manipulation.
    imagemagick_exe = "convert" if os.name != "nt" else "magick"

    rv = locate_executable(imagemagick_exe)
    if rv is not None:
        return rv

    # Give up.
    raise RuntimeError("Could not locate imagemagick.")


def get_thumbnail_ext(source_filename):
    ext = source_filename.rsplit(".", 1)[-1].lower()
    # if the extension is already of a format that a browser understands
    # we will roll with it.
    if ext.lower() in ("png", "jpg", "jpeg", "gif"):
        return None
    # Otherwise we roll with JPEG as default.
    return ".jpeg"


def get_quality(source_filename):
    ext = source_filename.rsplit(".", 1)[-1].lower()
    if ext.lower() == "png":
        return 75
    return 85


def compute_dimensions(width, height, source_width, source_height):
    """computes the bounding dimensions"""
    computed_width, computed_height = width, height

    width, height, source_width, source_height = (
        None if v is None else float(v)
        for v in (width, height, source_width, source_height)
    )

    source_ratio = source_width / source_height

    def _round(x):
        # make sure things get top-rounded, to be consistent with imagemagick
        return int(decimal.Decimal(x).to_integral(decimal.ROUND_HALF_UP))

    if width is None or (height is not None and width / height > source_ratio):
        computed_width = _round(height * source_ratio)
    else:
        computed_height = _round(width / source_ratio)

    return computed_width, computed_height


def process_image(
    ctx,
    source_image,
    dst_filename,
    width=None,
    height=None,
    mode=ThumbnailMode.DEFAULT,
    quality=None,
):
    """Build image from source image, optionally compressing and resizing.

    "source_image" is the absolute path of the source in the content directory,
    "dst_filename" is the absolute path of the target in the output directory.
    """
    if width is None and height is None:
        raise ValueError("Must specify at least one of width or height.")

    im = find_imagemagick(ctx.build_state.config["IMAGEMAGICK_EXECUTABLE"])

    if quality is None:
        quality = get_quality(source_image)

    resize_key = ""
    if width is not None:
        resize_key += str(width)
    if height is not None:
        resize_key += "x" + str(height)

    if mode == ThumbnailMode.STRETCH:
        resize_key += "!"

    cmdline = [im, source_image, "-auto-orient"]
    if mode == ThumbnailMode.CROP:
        cmdline += [
            "-resize",
            resize_key + "^",
            "-gravity",
            "Center",
            "-extent",
            resize_key,
        ]
    else:
        cmdline += ["-resize", resize_key]

    cmdline += ["-strip", "-colorspace", "sRGB"]
    cmdline += ["-quality", str(quality), dst_filename]

    reporter.report_debug_info("imagemagick cmd line", cmdline)
    portable_popen(cmdline).wait()


def make_image_thumbnail(
    ctx,
    source_image,
    source_url_path,
    width=None,
    height=None,
    mode=ThumbnailMode.DEFAULT,
    upscale=None,
    quality=None,
):
    """Helper method that can create thumbnails from within the build process
    of an artifact.
    """
    if width is None and height is None:
        raise ValueError("Must specify at least one of width or height.")

    # temporarily fallback to "fit" in case of erroneous arguments
    # to preserve backward-compatibility.
    # this needs to change to an exception in the future.
    if mode != ThumbnailMode.FIT and (width is None or height is None):
        warnings.warn(
            f'"{mode.value}" mode requires both `width` and `height` '
            'to be specified. Falling back to "fit" mode.'
        )
        mode = ThumbnailMode.FIT

    if upscale is None and mode in (ThumbnailMode.CROP, ThumbnailMode.STRETCH):
        upscale = True

    with open(source_image, "rb") as f:
        format, source_width, source_height = get_image_info(f)

    if format is None:
        raise RuntimeError("Cannot process unknown images")

    # If we are dealing with an actual svg image, we do not actually
    # resize anything, we just return it. This is not ideal but it's
    # better than outright failing.
    if format == "svg":
        return Thumbnail(source_url_path, width, height)

    if mode == ThumbnailMode.FIT:
        computed_width, computed_height = compute_dimensions(
            width, height, source_width, source_height
        )
    else:
        computed_width, computed_height = width, height

    would_upscale = computed_width > source_width or computed_height > source_height

    # this part needs to be removed once backward-compatibility period passes
    if would_upscale and upscale is None:
        warnings.warn(
            "Your image is being scaled up since the requested thumbnail "
            "size is larger than the source. This default will change "
            "in the future. If you want to preserve the current behaviour, "
            "use `upscale=True`."
        )
        upscale = True

    if would_upscale and not upscale:
        return Thumbnail(source_url_path, source_width, source_height)

    suffix = get_suffix(width, height, mode, quality=quality)
    dst_url_path = get_dependent_url(
        source_url_path, suffix, ext=get_thumbnail_ext(source_image)
    )

    def build_thumbnail_artifact(artifact):
        artifact.ensure_dir()
        process_image(
            ctx,
            source_image,
            artifact.dst_filename,
            width,
            height,
            mode,
            quality=quality,
        )

    ctx.sub_artifact(artifact_name=dst_url_path, sources=[source_image])(
        build_thumbnail_artifact
    )

    return Thumbnail(dst_url_path, computed_width, computed_height)


class Thumbnail:
    """Holds information about a thumbnail."""

    def __init__(self, url_path, width, height=None):
        #: the `width` of the thumbnail in pixels.
        self.width = width
        #: the `height` of the thumbnail in pixels.
        self.height = height
        #: the URL path of the image.
        self.url_path = url_path

    def __str__(self):
        return posixpath.basename(self.url_path)



from lektor.quickstart import get_default_author
from lektor.quickstart import get_default_author_email


def test_default_author(os_user):
    assert get_default_author() == "Lektor Test"


def test_default_author_email():
    assert isinstance(get_default_author_email(), str)



def test_basic_editor(scratch_tree):
    sess = scratch_tree.edit("/")

    assert sess.id == ""
    assert sess.path == "/"
    assert sess.record is not None

    assert sess["_model"] == "page"
    assert sess["title"] == "Index"
    assert sess["body"] == "*Hello World!*"

    sess["body"] = "A new body"
    sess.commit()

    assert sess.closed

    with open(sess.get_fs_path(), encoding="utf-8") as f:
        assert f.read().splitlines() == [
            "_model: page",
            "---",
            "title: Index",
            "---",
            "body: A new body",
        ]


def test_create_alt(scratch_tree, scratch_pad):
    sess = scratch_tree.edit("/", alt="de")

    assert sess.id == ""
    assert sess.path == "/"
    assert sess.record is not None

    assert sess["_model"] == "page"
    assert sess["title"] == "Index"
    assert sess["body"] == "*Hello World!*"

    sess["body"] = "Hallo Welt!"
    sess.commit()

    assert sess.closed

    # When we use the editor to change this, we only want the fields that
    # changed compared to the base to be included.
    with open(sess.get_fs_path(alt="de"), encoding="utf-8") as f:
        assert f.read().splitlines() == ["body: Hallo Welt!"]

    scratch_pad.cache.flush()
    item = scratch_pad.get("/", alt="de")
    assert item["_slug"] == ""
    assert item["title"] == "Index"
    assert item["body"].source == "Hallo Welt!"
    assert item["_model"] == "page"



import threading
from weakref import ref as weakref

import mistune
from markupsafe import escape
from markupsafe import Markup
from werkzeug.urls import url_parse

from lektor.context import get_ctx


_markdown_cache = threading.local()


class ImprovedRenderer(mistune.Renderer):
    def link(self, link, title, text):
        if self.record is not None:
            url = url_parse(link)
            if not url.scheme:
                link = self.record.url_to("!" + link, base_url=get_ctx().base_url)
        link = escape(link)
        if not title:
            return '<a href="%s">%s</a>' % (link, text)
        title = escape(title)
        return '<a href="%s" title="%s">%s</a>' % (link, title, text)

    def image(self, src, title, text):
        if self.record is not None:
            url = url_parse(src)
            if not url.scheme:
                src = self.record.url_to("!" + src, base_url=get_ctx().base_url)
        src = escape(src)
        text = escape(text)
        if title:
            title = escape(title)
            return '<img src="%s" alt="%s" title="%s">' % (src, text, title)
        return '<img src="%s" alt="%s">' % (src, text)


class MarkdownConfig:
    def __init__(self):
        self.options = {
            "escape": False,
        }
        self.renderer_base = ImprovedRenderer
        self.renderer_mixins = []

    def make_renderer(self):
        bases = tuple(self.renderer_mixins) + (self.renderer_base,)
        renderer_cls = type("renderer_cls", bases, {})
        return renderer_cls(**self.options)


def make_markdown(env):
    cfg = MarkdownConfig()
    env.plugin_controller.emit("markdown-config", config=cfg)
    renderer = cfg.make_renderer()
    env.plugin_controller.emit("markdown-lexer-config", config=cfg, renderer=renderer)
    return mistune.Markdown(renderer, **cfg.options)


def markdown_to_html(text, record=None):
    ctx = get_ctx()
    if ctx is None:
        raise RuntimeError("Context is required for markdown rendering")

    # These markdown parsers are all terrible.  Not one of them does not
    # modify internal state.  So since we only do one of those per thread
    # we can at least cache them on a thread local.
    md = getattr(_markdown_cache, "md", None)
    if md is None:
        md = make_markdown(ctx.env)
        _markdown_cache.md = md

    meta = {}
    ctx.env.plugin_controller.emit("markdown-meta-init", meta=meta, record=record)
    md.renderer.meta = meta
    md.renderer.record = record
    rv = md(text)
    ctx.env.plugin_controller.emit(
        "markdown-meta-postprocess", meta=meta, record=record
    )
    return rv, meta


class Markdown:
    def __init__(self, source, record=None):
        self.source = source
        self.__record = weakref(record) if record is not None else lambda: None
        self.__cached_for_ctx = None
        self.__html = None
        self.__meta = None

    def __bool__(self):
        return bool(self.source)

    __nonzero__ = __bool__

    def __render(self):
        # When the markdown instance is attached to a cached object we can
        # end up in the situation where the context changed from the time
        # we were put into the cache to the time where we got referenced
        # by something elsewhere.  In that case we need to re-process our
        # markdown.  For instance this affects relative links.
        if self.__html is None or self.__cached_for_ctx != get_ctx():
            self.__html, self.__meta = markdown_to_html(self.source, self.__record())
            self.__cached_for_ctx = get_ctx()

    @property
    def meta(self):
        self.__render()
        return self.__meta

    @property
    def html(self):
        self.__render()
        return Markup(self.__html)

    def __getitem__(self, name):
        return self.meta[name]

    def __str__(self):
        self.__render()
        return self.__html

    def __html__(self):
        self.__render()
        return Markup(self.__html)



import errno
import os

from inifile import IniFile

from lektor.constants import PRIMARY_ALT
from lektor.environment.expressions import Expression
from lektor.environment.expressions import FormatExpression
from lektor.i18n import generate_i18n_kvs
from lektor.i18n import get_i18n_block
from lektor.pagination import Pagination
from lektor.reporter import reporter
from lektor.types import builtin_types
from lektor.types.base import RawValue
from lektor.utils import bool_from_string
from lektor.utils import slugify


class ChildConfig:
    def __init__(
        self,
        enabled=None,
        slug_format=None,
        model=None,
        order_by=None,
        replaced_with=None,
        hidden=None,
    ):
        if enabled is None:
            enabled = True
        self.enabled = enabled
        self.slug_format = slug_format
        self.model = model
        self.order_by = order_by
        self.replaced_with = replaced_with
        self.hidden = hidden

    def to_json(self):
        return {
            "enabled": self.enabled,
            "slug_format": self.slug_format,
            "model": self.model,
            "order_by": self.order_by,
            "replaced_with": self.replaced_with,
            "hidden": self.hidden,
        }


class PaginationConfig:
    def __init__(self, env, enabled=None, per_page=None, url_suffix=None, items=None):
        self.env = env
        if enabled is None:
            enabled = False
        self.enabled = enabled
        if per_page is None:
            per_page = 20
        elif not isinstance(per_page, int):
            raise TypeError(f"per_page must be an int or None, not {per_page!r}")
        if per_page <= 0:
            raise ValueError("per_page must be positive, not {per_page}")

        self.per_page = per_page
        if url_suffix is None:
            url_suffix = "page"
        self.url_suffix = url_suffix
        self.items = items
        self._items_tmpl = None

    def count_total_items(self, record):
        """Counts the number of items over all pages."""
        return self.get_pagination_query(record).count()

    def count_pages(self, record):
        """Returns the total number of pages for the children of a record."""
        total = self.count_total_items(record)
        npages = (total + self.per_page - 1) // self.per_page
        # Even when there are no children, we want at least one page
        return max(npages, 1)

    def slice_query_for_page(self, record, page):
        """Slices the query so it returns the children for a given page."""
        query = self.get_pagination_query(record)
        if not self.enabled or page is None:
            return query
        return query.limit(self.per_page).offset((page - 1) * self.per_page)

    @staticmethod
    def get_record_for_page(record, page_num):
        """Given a normal record this one returns the version specific
        for a page.
        """
        # If we already have the right version, return it.
        if record.page_num == page_num:
            return record

        # Check if we have a cached version
        pad = record.pad
        rv = pad.cache.get(record.path, record.alt, str(page_num))
        if rv is not Ellipsis:
            return rv

        # Make what we need out of what we have and put it into the cache.
        cls = record.__class__
        rv = cls(record.pad, record._data, page_num=page_num)
        pad.cache.remember(rv)
        return rv

    def match_pagination(self, record, url_path):
        """Matches the pagination from the URL path."""
        if not self.enabled:
            return None
        suffixes = self.url_suffix.strip("/").split("/")
        if url_path[: len(suffixes)] != suffixes:
            return None
        try:
            page_num = int(url_path[len(suffixes)])
        except (ValueError, IndexError):
            return None

        # It's important we do not allow "1" here as the first page is always
        # on the root.  Changing this would mean the URLs are incorrectly
        # generated if someone manually went to /page/1/.
        if page_num == 1 or len(url_path) != len(suffixes) + 1:
            return None

        # Page needs to have at least a single child.
        rv = self.get_record_for_page(record, page_num)
        if rv.pagination.items.first() is not None:
            return rv
        return None

    def get_pagination_controller(self, record):
        if not self.enabled:
            raise RuntimeError("Pagination is disabled")
        return Pagination(record, self)

    def get_pagination_query(self, record):
        items_expr = self.items
        if items_expr is None:
            return record.children
        if self._items_tmpl is None or self._items_tmpl[0] != items_expr:
            self._items_tmpl = (items_expr, Expression(self.env, items_expr))

        return self._items_tmpl[1].evaluate(record.pad, this=record)

    def to_json(self):
        return {
            "enabled": self.enabled,
            "per_page": self.per_page,
            "url_suffix": self.url_suffix,
            "items": self.items,
        }


class AttachmentConfig:
    def __init__(self, enabled=None, model=None, order_by=None, hidden=None):
        if enabled is None:
            enabled = True
        if hidden is None:
            hidden = False
        self.enabled = enabled
        self.model = model
        self.order_by = order_by
        self.hidden = hidden

    def to_json(self):
        return {
            "enabled": self.enabled,
            "model": self.model,
            "order_by": self.order_by,
            "hidden": self.hidden,
        }


class Field:
    def __init__(self, env, name, type=None, options=None):
        if type is None:
            type = env.types["string"]
        if options is None:
            options = {}
        self.options = options
        self.name = name
        label_i18n = get_i18n_block(options, "label")
        if not label_i18n:
            label_i18n = {"en": name.replace("_", " ").strip().capitalize()}
        self.label_i18n = label_i18n
        self.description_i18n = get_i18n_block(options, "description") or None
        self.default = options.get("default")
        self.type = type(env, options)

    @property
    def label(self):
        return self.label_i18n.get("en")

    def to_json(self, pad, record=None, alt=PRIMARY_ALT):
        return {
            "name": self.name,
            "label": self.label,
            "label_i18n": self.label_i18n,
            "hide_label": bool_from_string(
                self.options.get("hide_label"), default=False
            ),
            "description_i18n": self.description_i18n,
            "type": self.type.to_json(pad, record, alt),
            "default": self.default,
            "alts_enabled": bool_from_string(
                self.options.get("alts_enabled"), default=None
            ),
        }

    def deserialize_value(self, value, pad=None):
        raw_value = RawValue(self.name, value, field=self, pad=pad)
        return self.type.value_from_raw_with_default(raw_value)

    def serialize_value(self, value):
        return self.type.value_to_raw(value)

    def __repr__(self):
        return "<%s %r type=%r>" % (
            self.__class__.__name__,
            self.name,
            self.type,
        )


def _iter_all_fields(obj):
    for name in sorted(x for x in obj.field_map if x[:1] == "_"):
        yield obj.field_map[name]
    for field in obj.fields:
        yield field


class DataModel:
    def __init__(
        self,
        env,
        id,
        name_i18n,
        label_i18n=None,
        filename=None,
        hidden=None,
        protected=None,
        child_config=None,
        attachment_config=None,
        pagination_config=None,
        fields=None,
        primary_field=None,
        parent=None,
    ):
        self.env = env
        self.filename = filename
        self.id = id
        self.name_i18n = name_i18n
        self.label_i18n = label_i18n
        if hidden is None:
            hidden = False
        self.hidden = hidden
        if protected is None:
            protected = False
        self.protected = protected
        if child_config is None:
            child_config = ChildConfig()
        self.child_config = child_config
        if attachment_config is None:
            attachment_config = AttachmentConfig()
        self.attachment_config = attachment_config
        if pagination_config is None:
            pagination_config = PaginationConfig(env)
        self.pagination_config = pagination_config
        if fields is None:
            fields = []
        self.fields = fields
        if primary_field is None and fields:
            primary_field = fields[0].name
        self.primary_field = primary_field
        self.parent = parent

        # This is a mapping of the key names to the actual field which
        # also includes the system fields.  This is primarily used for
        # fast internal operations but also the admin.
        self.field_map = dict((x.name, x) for x in fields)
        for key, (ty, opts) in system_fields.items():
            self.field_map[key] = Field(env, name=key, type=ty, options=opts)

        self._child_slug_tmpl = None
        self._child_replacements = None
        self._label_tmpls = {}

    @property
    def name(self):
        name = (self.name_i18n or {}).get("en")
        return name or self.id.title().replace("_", " ")

    @property
    def label(self):
        return (self.label_i18n or {}).get("en")

    def to_json(self, pad, record=None, alt=PRIMARY_ALT):
        """Describes the datamodel as JSON data."""
        return {
            "filename": self.filename,
            "alt": alt,
            "id": self.id,
            "name": self.name,
            "name_i18n": self.name_i18n,
            "primary_field": self.primary_field,
            "label": self.label,
            "label_i18n": self.label_i18n,
            "hidden": self.hidden,
            "protected": self.protected,
            "child_config": self.child_config.to_json(),
            "attachment_config": self.attachment_config.to_json(),
            "pagination_config": self.pagination_config.to_json(),
            "fields": [x.to_json(pad, record, alt) for x in _iter_all_fields(self)],
        }

    def format_record_label(self, record, lang="en"):
        """Returns the label for a given record."""
        label = self.label_i18n.get(lang)
        if label is None:
            return None

        tmpl = self._label_tmpls.get(lang)
        if tmpl is None:
            tmpl = (label, FormatExpression(self.env, label))
            self._label_tmpls[lang] = tmpl

        try:
            return tmpl[1].evaluate(record.pad, this=record)
        except Exception:
            # XXX: log
            return None

    def get_default_child_slug(self, pad, data):
        """Formats out the child slug."""
        slug_format = self.child_config.slug_format
        if slug_format is None:
            return data["_id"]

        if self._child_slug_tmpl is None or self._child_slug_tmpl[0] != slug_format:
            self._child_slug_tmpl = (
                slug_format,
                FormatExpression(self.env, slug_format),
            )

        try:
            return "_".join(
                self._child_slug_tmpl[1].evaluate(pad, this=data).strip().split()
            ).strip("/")
        except Exception as exc:
            reporter.report_generic("Failed to expand child slug_format: %s" % exc)
            return "temp-" + slugify(data["_id"])

    def get_default_template_name(self):
        return self.id + ".html"

    @property
    def has_own_children(self):
        return self.child_config.replaced_with is None and self.child_config.enabled

    @property
    def has_own_attachments(self):
        return self.attachment_config.enabled

    def get_child_replacements(self, record):
        """Returns the query that should be used as replacement for the
        actual children.
        """
        replaced_with = self.child_config.replaced_with
        if replaced_with is None:
            return None

        if (
            self._child_replacements is None
            or self._child_replacements[0] != replaced_with
        ):
            self._child_replacements = (
                replaced_with,
                Expression(self.env, replaced_with),
            )

        return self._child_replacements[1].evaluate(record.pad, this=record)

    def process_raw_data(self, raw_data, pad=None):
        rv = {}
        for field in self.field_map.values():
            value = raw_data.get(field.name)
            rv[field.name] = field.deserialize_value(value, pad=pad)
        rv["_model"] = self.id
        return rv

    def __repr__(self):
        return "<%s %r>" % (
            self.__class__.__name__,
            self.id,
        )


class FlowBlockModel:
    def __init__(
        self,
        env,
        id,
        name_i18n,
        filename=None,
        fields=None,
        order=None,
        button_label=None,
    ):
        self.env = env
        self.id = id
        self.name_i18n = name_i18n
        self.filename = filename
        if fields is None:
            fields = []
        self.fields = fields
        if order is None:
            order = 100
        self.order = order
        self.button_label = button_label

        self.field_map = dict((x.name, x) for x in fields)
        self.field_map["_flowblock"] = Field(
            env, name="_flowblock", type=env.types["string"]
        )

    @property
    def name(self):
        return self.name_i18n.get("en") or self.id.title().replace("_", " ")

    def to_json(self, pad, record=None, alt=PRIMARY_ALT):
        return {
            "id": self.id,
            "name": self.name,
            "name_i18n": self.name_i18n,
            "filename": self.filename,
            "fields": [
                x.to_json(pad, record, alt)
                for x in _iter_all_fields(self)
                if x.name != "_flowblock"
            ],
            "order": self.order,
            "button_label": self.button_label,
        }

    def process_raw_data(self, raw_data, pad=None):
        rv = {}
        for field in self.field_map.values():
            value = raw_data.get(field.name)
            rv[field.name] = field.deserialize_value(value, pad=pad)
        rv["_flowblock"] = self.id
        return rv

    def __repr__(self):
        return "<%s %r>" % (
            self.__class__.__name__,
            self.id,
        )


def fielddata_from_ini(inifile):
    return [
        (
            sect.split(".", 1)[1],
            inifile.section_as_dict(sect),
        )
        for sect in inifile.sections()
        if sect.startswith("fields.")
    ]


def datamodel_data_from_ini(id, inifile):
    def _parse_order(value):
        value = (value or "").strip()
        if not value:
            return None
        return [x for x in [x.strip() for x in value.strip().split(",")] if x]

    return dict(
        filename=inifile.filename,
        id=id,
        parent=inifile.get("model.inherits"),
        name_i18n=get_i18n_block(inifile, "model.name"),
        label_i18n=get_i18n_block(inifile, "model.label"),
        primary_field=inifile.get("model.primary_field"),
        hidden=inifile.get_bool("model.hidden", default=None),
        protected=inifile.get_bool("model.protected", default=None),
        child_config=dict(
            enabled=inifile.get_bool("children.enabled", default=None),
            slug_format=inifile.get("children.slug_format"),
            model=inifile.get("children.model"),
            order_by=_parse_order(inifile.get("children.order_by")),
            replaced_with=inifile.get("children.replaced_with"),
            hidden=inifile.get_bool("children.hidden", default=None),
        ),
        attachment_config=dict(
            enabled=inifile.get_bool("attachments.enabled", default=None),
            model=inifile.get("attachments.model"),
            order_by=_parse_order(inifile.get("attachments.order_by")),
            hidden=inifile.get_bool("attachments.hidden", default=None),
        ),
        pagination_config=dict(
            enabled=inifile.get_bool("pagination.enabled", default=None),
            per_page=inifile.get_int("pagination.per_page"),
            url_suffix=inifile.get("pagination.url_suffix"),
            items=inifile.get("pagination.items"),
        ),
        fields=fielddata_from_ini(inifile),
    )


def flowblock_data_from_ini(id, inifile):
    return dict(
        filename=inifile.filename,
        id=id,
        name_i18n=get_i18n_block(inifile, "block.name"),
        fields=fielddata_from_ini(inifile),
        order=inifile.get_int("block.order"),
        button_label=inifile.get("block.button_label"),
    )


def fields_from_data(env, data, parent_fields=None):
    fields = []
    known_fields = set()

    for name, options in data:
        ty = env.types[options.get("type", "string")]
        fields.append(Field(env=env, name=name, type=ty, options=options))
        known_fields.add(name)

    if parent_fields is not None:
        prepended_fields = []
        for field in parent_fields:
            if field.name not in known_fields:
                prepended_fields.append(field)
        fields = prepended_fields + fields

    return fields


def datamodel_from_data(env, model_data, parent=None):
    def get_value(key):
        path = key.split(".")
        node = model_data
        for item in path:
            node = node.get(item)
        if node is not None:
            return node
        if parent is not None:
            node = parent
            for item in path:
                node = getattr(node, item)
            return node
        return None

    fields = fields_from_data(
        env, model_data["fields"], parent and parent.fields or None
    )

    return DataModel(
        env,
        # data that never inherits
        filename=model_data["filename"],
        id=model_data["id"],
        parent=parent,
        name_i18n=model_data["name_i18n"],
        primary_field=model_data["primary_field"],
        # direct data that can inherit
        label_i18n=get_value("label_i18n"),
        hidden=get_value("hidden"),
        protected=get_value("protected"),
        child_config=ChildConfig(
            enabled=get_value("child_config.enabled"),
            slug_format=get_value("child_config.slug_format"),
            model=get_value("child_config.model"),
            order_by=get_value("child_config.order_by"),
            replaced_with=get_value("child_config.replaced_with"),
            hidden=get_value("child_config.hidden"),
        ),
        attachment_config=AttachmentConfig(
            enabled=get_value("attachment_config.enabled"),
            model=get_value("attachment_config.model"),
            order_by=get_value("attachment_config.order_by"),
            hidden=get_value("attachment_config.hidden"),
        ),
        pagination_config=PaginationConfig(
            env,
            enabled=get_value("pagination_config.enabled"),
            per_page=get_value("pagination_config.per_page"),
            url_suffix=get_value("pagination_config.url_suffix"),
            items=get_value("pagination_config.items"),
        ),
        fields=fields,
    )


def flowblock_from_data(env, block_data):
    return FlowBlockModel(
        env,
        filename=block_data["filename"],
        id=block_data["id"],
        name_i18n=block_data["name_i18n"],
        fields=fields_from_data(env, block_data["fields"]),
        order=block_data["order"],
        button_label=block_data["button_label"],
    )


def iter_inis(path):
    try:
        for filename in os.listdir(path):
            if not filename.endswith(".ini") or filename[:1] in "_.":
                continue
            fn = os.path.join(path, filename)
            if os.path.isfile(fn):
                base = filename[:-4]
                base = base.encode("utf-8").decode("ascii", "replace")
                inifile = IniFile(fn)
                yield base, inifile
    except OSError as e:
        if e.errno != errno.ENOENT:
            raise


def load_datamodels(env):
    """Loads the datamodels for a specific environment."""
    # Models will override previous loaded models with the same name
    # So models paths are loaded in reverse order
    paths = list(reversed(env.theme_paths)) + [env.root_path]
    paths = [os.path.join(p, "models") for p in paths]
    data = {}

    for path in paths:
        for model_id, inifile in iter_inis(path):
            data[model_id] = datamodel_data_from_ini(model_id, inifile)

    rv = {}

    def get_model(model_id):
        model = rv.get(model_id)
        if model is not None:
            return model
        if model_id in data:
            return create_model(model_id)
        return None

    def create_model(model_id):
        model_data = data.get(model_id)
        if model_data is None:
            raise RuntimeError("Model %r not found" % model_id)

        if model_data["parent"] is not None:
            parent = get_model(model_data["parent"])
        else:
            parent = None

        rv[model_id] = mod = datamodel_from_data(env, model_data, parent)
        return mod

    for model_id in data:
        get_model(model_id)

    rv["none"] = DataModel(env, "none", {"en": "None"}, hidden=True)

    return rv


def load_flowblocks(env):
    """Loads all the flow blocks for a specific environment."""
    # Flowblocks will override previous loaded flowblocks with the same name
    # So paths are loaded in reverse order
    paths = list(reversed(env.theme_paths)) + [env.root_path]
    paths = [os.path.join(p, "flowblocks") for p in paths]
    rv = {}

    for path in paths:
        for flowblock_id, inifile in iter_inis(path):
            rv[flowblock_id] = flowblock_from_data(
                env, flowblock_data_from_ini(flowblock_id, inifile)
            )

    return rv


system_fields = {}


def add_system_field(name, **opts):
    opts = dict(generate_i18n_kvs(**opts))
    ty = builtin_types[opts.pop("type")]
    system_fields[name] = (ty, opts)


# The full path of the record
add_system_field("_path", type="string")

# The local ID (within a folder) of the record
add_system_field("_id", type="string")

# The global ID (within a folder) of the record
add_system_field("_gid", type="string")

# The alt key that identifies this record
add_system_field("_alt", type="string")

# The alt key for the file that was actually referenced.
add_system_field("_source_alt", type="string")

# the model that defines the data of the record
add_system_field("_model", type="string")

# the template that should be used for rendering if not hidden
add_system_field(
    "_template",
    type="string",
    label_i18n="TEMPLATE",
    width="1/2",
    addon_label="[[code]]",
)

# the slug that should be used for this record.  This is added below the
# slug of the parent.
add_system_field("_slug", type="slug", label_i18n="URL_SLUG", width="1/2")

# This can be used to hide an individual record.
add_system_field(
    "_hidden",
    type="boolean",
    label_i18n="HIDE_PAGE",
    checkbox_label_i18n="HIDE_PAGE_EXPLANATION",
)

# This marks a page as undiscoverable.
add_system_field(
    "_discoverable",
    type="boolean",
    default="yes",
    label_i18n="PAGE_IS_DISCOVERABLE",
    checkbox_label_i18n="PAGE_IS_DISCOVERABLE_EXPLANATION",
)

# Useful fields for attachments.
add_system_field("_attachment_for", type="string")
add_system_field(
    "_attachment_type",
    type="string",
    label_i18n="ATTACHMENT_TYPE",
    addon_label="[[paperclip]]",
)



import mimetypes
import os
import posixpath
from io import BytesIO
from zlib import adler32

from flask import abort
from flask import Blueprint
from flask import current_app
from flask import render_template
from flask import request
from flask import Response
from werkzeug.datastructures import Headers
from werkzeug.exceptions import NotFound
from werkzeug.wsgi import wrap_file


bp = Blueprint("serve", __name__)


def rewrite_html_for_editing(fp, edit_url):
    contents = fp.read()

    button_style = u"""
    <style type="text/css">
      #lektor-edit-link {
        position: fixed;
        z-index: 9999999;
        right: 10px;
        top: 10px;
        position: fixed;
        margin: 0;
        font-family: 'Verdana', sans-serif;
        background: #eee;
        color: #77304c;
        font-weight: normal;
        font-size: 32px;
        padding: 0;
        text-decoration: none!important;
        border: 1px solid #ccc!important;
        width: 40px;
        height: 40px;
        line-height: 40px;
        text-align: center;
        opacity: 0.7;
      }

      #lektor-edit-link:hover {
        background: white!important;
        opacity: 1.0;
        border: 1px solid #aaa!important;
      }
    </style>
    """
    button_script = u"""
    <script type="text/javascript">
      (function() {
        if (window != window.top) {
          return;
        }
        var link = document.createElement('a');
        link.setAttribute('href', '%(edit_url)s?path=' +
            encodeURIComponent(document.location.pathname));
        link.setAttribute('id', 'lektor-edit-link');
        link.innerHTML = '\u270E';
        document.body.appendChild(link);
      })();
    </script>
    """ % {
        "edit_url": edit_url,
    }

    head_endpos = contents.find(b"</head>")
    body_endpos = contents.find(b"</body>")
    if head_endpos >= 0 and body_endpos >= 0:
        return BytesIO(
            contents[:head_endpos]
            + button_style.encode("utf-8")
            + contents[head_endpos:body_endpos]
            + button_script.encode("utf-8")
            + contents[body_endpos:]
        )

    return BytesIO(
        contents + button_style.encode("utf-8") + button_script.encode("utf-8")
    )


def send_file(filename):
    # pylint: disable=consider-using-with
    mimetype = mimetypes.guess_type(filename)[0]
    if mimetype is None:
        mimetype = "application/octet-stream"

    headers = Headers()

    try:
        file = open(filename, "rb")
        mtime = os.path.getmtime(filename)
        headers["Content-Length"] = os.path.getsize(filename)
    except (IOError, OSError):
        abort(404)

    rewritten = False
    if mimetype == "text/html":
        rewritten = True
        new_file = rewrite_html_for_editing(
            file, edit_url=posixpath.join("/", request.script_root, "admin/edit")
        )
        file.close()
        file = new_file
        del headers["Content-Length"]

    headers["Cache-Control"] = "no-cache, no-store"

    data = wrap_file(request.environ, file)

    rv = Response(data, mimetype=mimetype, headers=headers, direct_passthrough=True)

    if not rewritten:
        # if we know the file modification date, we can store it as
        # the time of the last modification.
        if mtime is not None:
            rv.last_modified = int(mtime)
        rv.cache_control.public = True
        try:
            rv.set_etag(
                "lektor-%s-%s-%s"
                % (
                    os.path.getmtime(filename),
                    os.path.getsize(filename),
                    adler32(
                        filename.encode("utf-8")
                        if isinstance(filename, str)
                        else filename
                    )
                    & 0xFFFFFFFF,
                )
            )
        except OSError:
            pass

    return rv


def handle_build_failure(failure):
    return render_template("build-failure.html", **failure.data)


def serve_up_artifact(path):
    li = current_app.lektor_info
    pad = li.get_pad()

    artifact_name, filename = li.resolve_artifact("/" + path, pad)
    if filename is None:
        abort(404)

    if artifact_name is None:
        artifact_name = path.strip("/")

    # If there was a build failure for the given artifact, we want
    # to render this instead of sending the (most likely missing or
    # corrupted) file.
    ctrl = li.get_failure_controller(pad)
    failure = ctrl.lookup_failure(artifact_name)
    if failure is not None:
        return handle_build_failure(failure)

    return send_file(filename)


@bp.route("/", defaults={"path": ""})
@bp.route("/<path:path>")
def serve_artifact(path):
    return serve_up_artifact(path)


@bp.errorhandler(404)
def serve_error_page(error):
    try:
        return serve_up_artifact("404.html")
    except NotFound as e:
        return e



import copy
import os
import re
from collections import OrderedDict

from inifile import IniFile
from werkzeug.urls import url_parse
from werkzeug.utils import cached_property

from lektor.constants import PRIMARY_ALT
from lektor.i18n import get_i18n_block
from lektor.utils import bool_from_string
from lektor.utils import secure_url


DEFAULT_CONFIG = {
    "IMAGEMAGICK_EXECUTABLE": None,
    "EPHEMERAL_RECORD_CACHE_SIZE": 500,
    "ATTACHMENT_TYPES": {
        # Only enable image formats here that we can handle in imagetools.
        # Right now this is limited to jpg, png and gif because this is
        # the only thing we compile into imagemagick on OS X distributions
        # as those are what browsers also support.  Thers is no point in
        # adding others here as we do not force convert images (yet?) but
        # only use it for thumbnailing.  However an image should be
        # visible even without thumbnailing.
        ".jpg": "image",
        ".jpeg": "image",
        ".png": "image",
        ".gif": "image",
        ".svg": "image",
        ".avi": "video",
        ".mpg": "video",
        ".mpeg": "video",
        ".wmv": "video",
        ".ogv": "video",
        ".mp4": "video",
        ".mp3": "audio",
        ".wav": "audio",
        ".ogg": "audio",
        ".pdf": "document",
        ".doc": "document",
        ".docx": "document",
        ".htm": "document",
        ".html": "document",
        ".txt": "text",
        ".log": "text",
    },
    "PROJECT": {
        "name": None,
        "locale": "en_US",
        "url": None,
        "url_style": "relative",
    },
    "THEME_SETTINGS": {},
    "PACKAGES": {},
    "ALTERNATIVES": OrderedDict(),
    "PRIMARY_ALTERNATIVE": None,
    "SERVERS": {},
}


def update_config_from_ini(config, inifile):
    def set_simple(target, source_path):
        rv = config.get(source_path)
        if rv is not None:
            config[target] = rv

    set_simple(
        target="IMAGEMAGICK_EXECUTABLE", source_path="env.imagemagick_executable"
    )
    set_simple(target="LESSC_EXECUTABLE", source_path="env.lessc_executable")

    for section_name in ("ATTACHMENT_TYPES", "PROJECT", "PACKAGES", "THEME_SETTINGS"):
        section_config = inifile.section_as_dict(section_name.lower())
        config[section_name].update(section_config)

    for sect in inifile.sections():
        if sect.startswith("servers."):
            server_id = sect.split(".")[1]
            config["SERVERS"][server_id] = inifile.section_as_dict(sect)
        elif sect.startswith("alternatives."):
            alt = sect.split(".")[1]
            config["ALTERNATIVES"][alt] = {
                "name": get_i18n_block(inifile, "alternatives.%s.name" % alt),
                "url_prefix": inifile.get("alternatives.%s.url_prefix" % alt),
                "url_suffix": inifile.get("alternatives.%s.url_suffix" % alt),
                "primary": inifile.get_bool("alternatives.%s.primary" % alt),
                "locale": inifile.get("alternatives.%s.locale" % alt, "en_US"),
            }

    for alt, alt_data in config["ALTERNATIVES"].items():
        if alt_data["primary"]:
            config["PRIMARY_ALTERNATIVE"] = alt
            break
    else:
        if config["ALTERNATIVES"]:
            raise RuntimeError("Alternatives defined but no primary set.")


class ServerInfo:
    def __init__(self, id, name_i18n, target, enabled=True, default=False, extra=None):
        self.id = id
        self.name_i18n = name_i18n
        self.target = target
        self.enabled = enabled
        self.default = default
        self.extra = extra or {}

    @property
    def name(self):
        return self.name_i18n.get("en") or self.id

    @property
    def short_target(self):
        match = re.match(r"([a-z]+)://([^/]+)", self.target)
        if match is not None:
            protocol, server = match.groups()
            return "%s via %s" % (server, protocol)
        return self.target

    def to_json(self):
        return {
            "id": self.id,
            "name": self.name,
            "name_i18n": self.name_i18n,
            "target": self.target,
            "short_target": self.short_target,
            "enabled": self.enabled,
            "default": self.default,
            "extra": self.extra,
        }


class Config:
    def __init__(self, filename=None):
        self.filename = filename
        self.values = copy.deepcopy(DEFAULT_CONFIG)

        if filename is not None and os.path.isfile(filename):
            inifile = IniFile(filename)
            update_config_from_ini(self.values, inifile)

    def __getitem__(self, name):
        return self.values[name]

    @property
    def site_locale(self):
        """The locale of this project."""
        return self.values["PROJECT"]["locale"]

    def get_servers(self, public=False):
        """Returns a list of servers."""
        rv = {}
        for server in self.values["SERVERS"]:
            server_info = self.get_server(server, public=public)
            if server_info is None:
                continue
            rv[server] = server_info
        return rv

    def get_default_server(self, public=False):
        """Returns the default server."""
        choices = []
        for server in self.values["SERVERS"]:
            server_info = self.get_server(server, public=public)
            if server_info is not None:
                if server_info.default:
                    return server_info
                choices.append(server_info)
        if len(choices) == 1:
            return choices[0]
        return None

    def get_server(self, name, public=False):
        """Looks up a server info by name."""
        info = self.values["SERVERS"].get(name)
        if info is None or "target" not in info:
            return None
        info = info.copy()
        target = info.pop("target")
        if public:
            target = secure_url(target)
        return ServerInfo(
            id=name,
            name_i18n=get_i18n_block(info, "name", pop=True),
            target=target,
            enabled=bool_from_string(info.pop("enabled", None), True),
            default=bool_from_string(info.pop("default", None), False),
            extra=info,
        )

    def is_valid_alternative(self, alt):
        """Checks if an alternative ID is known."""
        if alt == PRIMARY_ALT:
            return True
        return alt in self.values["ALTERNATIVES"]

    def list_alternatives(self):
        """Returns a sorted list of alternative IDs."""
        return sorted(self.values["ALTERNATIVES"])

    def iter_alternatives(self):
        """Iterates over all alterantives.  If the system is disabled this
        yields '_primary'.
        """
        found = False
        for alt in self.values["ALTERNATIVES"]:
            if alt != PRIMARY_ALT:
                yield alt
                found = True
        if not found:
            yield PRIMARY_ALT

    def get_alternative(self, alt):
        """Returns the config setting of the given alt."""
        if alt == PRIMARY_ALT:
            alt = self.primary_alternative
        return self.values["ALTERNATIVES"].get(alt)

    def get_alternative_url_prefixes(self):
        """Returns a list of alternative url prefixes by length."""
        items = [
            (v["url_prefix"].lstrip("/"), k)
            for k, v in self.values["ALTERNATIVES"].items()
            if v["url_prefix"]
        ]
        items.sort(key=lambda x: -len(x[0]))
        return items

    def get_alternative_url_suffixes(self):
        """Returns a list of alternative url suffixes by length."""
        items = [
            (v["url_suffix"].rstrip("/"), k)
            for k, v in self.values["ALTERNATIVES"].items()
            if v["url_suffix"]
        ]
        items.sort(key=lambda x: -len(x[0]))
        return items

    def get_alternative_url_span(self, alt=PRIMARY_ALT):
        """Returns the URL span (prefix, suffix) for an alt."""
        if alt == PRIMARY_ALT:
            alt = self.primary_alternative
        cfg = self.values["ALTERNATIVES"].get(alt)
        if cfg is not None:
            return cfg["url_prefix"] or "", cfg["url_suffix"] or ""
        return "", ""

    @cached_property
    def primary_alternative_is_rooted(self):
        """`True` if the primary alternative is sitting at the root of
        the URL handler.
        """
        primary = self.primary_alternative
        if primary is None:
            return True

        cfg = self.values["ALTERNATIVES"].get(primary)
        if not (cfg["url_prefix"] or "").lstrip("/") and not (
            cfg["url_suffix"] or ""
        ).rstrip("/"):
            return True

        return False

    @property
    def primary_alternative(self):
        """The identifier that acts as primary alternative."""
        return self.values["PRIMARY_ALTERNATIVE"]

    @cached_property
    def base_url(self):
        """The external base URL."""
        url = self.values["PROJECT"].get("url")
        if url and url_parse(url).scheme:
            return url.rstrip("/") + "/"
        return None

    @cached_property
    def base_path(self):
        """The base path of the URL."""
        url = self.values["PROJECT"].get("url")
        if url:
            return url_parse(url).path.rstrip("/") + "/"
        return "/"

    @cached_property
    def url_style(self):
        """The intended URL style."""
        style = self.values["PROJECT"].get("url_style")
        if style in ("relative", "absolute", "external"):
            return style
        return "relative"



from lektor.types.base import BadValue  # noqa - reexport
from lektor.types.base import get_undefined_info  # noqa - reexport
from lektor.types.base import RawValue  # noqa - reexport
from lektor.types.base import Type  # noqa - reexport
from lektor.types.fake import HeadingType
from lektor.types.fake import InfoType
from lektor.types.fake import LineType
from lektor.types.fake import SpacingType
from lektor.types.flow import FlowType
from lektor.types.formats import MarkdownType
from lektor.types.multi import CheckboxesType
from lektor.types.multi import SelectType
from lektor.types.primitives import BooleanType
from lektor.types.primitives import DateTimeType
from lektor.types.primitives import DateType
from lektor.types.primitives import FloatType
from lektor.types.primitives import HtmlType
from lektor.types.primitives import IntegerType
from lektor.types.primitives import StringsType
from lektor.types.primitives import StringType
from lektor.types.primitives import TextType
from lektor.types.special import SlugType
from lektor.types.special import SortKeyType
from lektor.types.special import UrlType


builtin_types = {
    # Primitive
    "string": StringType,
    "strings": StringsType,
    "text": TextType,
    "html": HtmlType,
    "integer": IntegerType,
    "float": FloatType,
    "boolean": BooleanType,
    "date": DateType,
    "datetime": DateTimeType,
    # Multi
    "checkboxes": CheckboxesType,
    "select": SelectType,
    # Special
    "sort_key": SortKeyType,
    "slug": SlugType,
    "url": UrlType,
    # Formats
    "markdown": MarkdownType,
    # Flow
    "flow": FlowType,
    # Fake
    "line": LineType,
    "spacing": SpacingType,
    "info": InfoType,
    "heading": HeadingType,
}



import hashlib
import os
import shutil
import sqlite3
import stat
import sys
import tempfile
from collections import deque
from collections import namedtuple
from contextlib import contextmanager
from itertools import chain

import click

from lektor.build_programs import builtin_build_programs
from lektor.buildfailures import FailureController
from lektor.constants import PRIMARY_ALT
from lektor.context import Context
from lektor.reporter import reporter
from lektor.sourcesearch import find_files
from lektor.utils import fs_enc
from lektor.utils import process_extra_flags
from lektor.utils import prune_file_and_folder


def create_tables(con):
    can_disable_rowid = ("3", "8") <= tuple(sqlite3.sqlite_version.split("."))
    if can_disable_rowid:
        without_rowid = "without rowid"
    else:
        without_rowid = ""

    try:
        con.execute(
            """
            create table if not exists artifacts (
                artifact text,
                source text,
                source_mtime integer,
                source_size integer,
                source_checksum text,
                is_dir integer,
                is_primary_source integer,
                primary key (artifact, source)
            ) %s;
        """
            % without_rowid
        )
        con.execute(
            """
            create index if not exists artifacts_source on artifacts (
                source
            );
        """
        )
        con.execute(
            """
            create table if not exists artifact_config_hashes (
                artifact text,
                config_hash text,
                primary key (artifact)
            ) %s;
        """
            % without_rowid
        )
        con.execute(
            """
            create table if not exists dirty_sources (
                source text,
                primary key (source)
            ) %s;
        """
            % without_rowid
        )
        con.execute(
            """
            create table if not exists source_info (
                path text,
                alt text,
                lang text,
                type text,
                source text,
                title text,
                primary key (path, alt, lang)
            ) %s;
        """
            % without_rowid
        )
    finally:
        con.close()


class BuildState:
    def __init__(self, builder, path_cache):
        self.builder = builder

        self.named_temporaries = set()
        self.updated_artifacts = []
        self.failed_artifacts = []
        self.path_cache = path_cache

    @property
    def pad(self):
        """The pad for this buildstate."""
        return self.builder.pad

    @property
    def env(self):
        """The environment backing this buildstate."""
        return self.builder.env

    @property
    def config(self):
        """The config for this buildstate."""
        return self.builder.pad.db.config

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, tb):
        self.close()

    def close(self):
        for fn in self.named_temporaries:
            try:
                os.remove(fn)
            except OSError:
                pass

    def notify_failure(self, artifact, exc_info):
        """Notify about a failure.  This marks a failed artifact and stores
        a failure.
        """
        self.failed_artifacts.append(artifact)
        self.builder.failure_controller.store_failure(artifact.artifact_name, exc_info)
        reporter.report_failure(artifact, exc_info)

    def make_named_temporary(self, identifier=None):
        """Creates a named temporary file and returns the filename for it.
        This can be usedful in some scenarious when building with external
        tools.
        """
        tmpdir = os.path.join(self.builder.meta_path, "tmp")
        try:
            os.makedirs(tmpdir)
        except OSError:
            pass
        fn = os.path.join(
            dir,
            "nt-%s-%s.tmp" % (identifier or "generic", os.urandom(20).encode("hex")),
        )
        self.named_temporaries.add(fn)
        return fn

    def get_file_info(self, filename):
        if filename:
            return self.path_cache.get_file_info(filename)
        return None

    def to_source_filename(self, filename):
        return self.path_cache.to_source_filename(filename)

    def get_virtual_source_info(self, virtual_source_path):
        virtual_source = self.pad.get(virtual_source_path)
        if not virtual_source:
            return VirtualSourceInfo(virtual_source_path, None, None)
        mtime = virtual_source.get_mtime(self.path_cache)
        checksum = virtual_source.get_checksum(self.path_cache)
        return VirtualSourceInfo(virtual_source_path, mtime, checksum)

    def connect_to_database(self):
        """Returns a database connection for the build state db."""
        return self.builder.connect_to_database()

    def get_destination_filename(self, artifact_name):
        """Returns the destination filename for an artifact name."""
        return os.path.join(
            self.builder.destination_path,
            artifact_name.strip("/").replace("/", os.path.sep),
        )

    def artifact_name_from_destination_filename(self, filename):
        """Returns the artifact name for a destination filename."""
        dst = self.builder.destination_path
        filename = os.path.join(dst, filename)
        if filename.startswith(dst):
            filename = filename[len(dst) :].lstrip(os.path.sep)
            if os.path.altsep:
                filename = filename.lstrip(os.path.altsep)
        return filename.replace(os.path.sep, "/")

    def new_artifact(
        self, artifact_name, sources=None, source_obj=None, extra=None, config_hash=None
    ):
        """Creates a new artifact and returns it."""
        dst_filename = self.get_destination_filename(artifact_name)
        key = self.artifact_name_from_destination_filename(dst_filename)
        return Artifact(
            self,
            key,
            dst_filename,
            sources,
            source_obj=source_obj,
            extra=extra,
            config_hash=config_hash,
        )

    def artifact_exists(self, artifact_name):
        """Given an artifact name this checks if it was already produced."""
        dst_filename = self.get_destination_filename(artifact_name)
        return os.path.exists(dst_filename)

    def get_artifact_dependency_infos(self, artifact_name, sources):
        con = self.connect_to_database()
        try:
            cur = con.cursor()
            rv = list(self._iter_artifact_dependency_infos(cur, artifact_name, sources))
        finally:
            con.close()
        return rv

    def _iter_artifact_dependency_infos(self, cur, artifact_name, sources):
        """This iterates over all dependencies as file info objects."""
        cur.execute(
            """
            select source, source_mtime, source_size,
                   source_checksum, is_dir
            from artifacts
            where artifact = ?
        """,
            [artifact_name],
        )
        rv = cur.fetchall()

        found = set()
        for path, mtime, size, checksum, is_dir in rv:
            if "@" in path:
                yield path, VirtualSourceInfo(path, mtime, checksum)
            else:
                file_info = FileInfo(
                    self.env, path, mtime, size, checksum, bool(is_dir)
                )
                filename = self.to_source_filename(file_info.filename)
                found.add(filename)
                yield filename, file_info

        # In any case we also iterate over our direct sources, even if the
        # build state does not know about them yet.  This can be caused by
        # an initial build or a change in original configuration.
        for source in sources:
            filename = self.to_source_filename(source)
            if filename not in found:
                yield source, None

    def write_source_info(self, info):
        """Writes the source info into the database.  The source info is
        an instance of :class:`lektor.build_programs.SourceInfo`.
        """
        reporter.report_write_source_info(info)
        source = self.to_source_filename(info.filename)
        con = self.connect_to_database()
        try:
            cur = con.cursor()
            for lang, title in info.title_i18n.items():
                cur.execute(
                    """
                    insert or replace into source_info
                        (path, alt, lang, type, source, title)
                        values (?, ?, ?, ?, ?, ?)
                """,
                    [info.path, info.alt, lang, info.type, source, title],
                )
            con.commit()
        finally:
            con.close()

    def prune_source_infos(self):
        """Remove all source infos of files that no longer exist."""
        MAX_VARS = 999  # Default SQLITE_MAX_VARIABLE_NUMBER.
        con = self.connect_to_database()
        to_clean = []
        try:
            cur = con.cursor()
            cur.execute(
                """
                select distinct source from source_info
            """
            )
            for (source,) in cur.fetchall():
                fs_path = os.path.join(self.env.root_path, source)
                if not os.path.exists(fs_path):
                    to_clean.append(source)

            if to_clean:
                for i in range(0, len(to_clean), MAX_VARS):
                    chunk = to_clean[i : i + MAX_VARS]
                    cur.execute(
                        """
                        delete from source_info
                         where source in (%s)
                    """
                        % ", ".join(["?"] * len(chunk)),
                        chunk,
                    )

                con.commit()
        finally:
            con.close()

        for source in to_clean:
            reporter.report_prune_source_info(source)

    def remove_artifact(self, artifact_name):
        """Removes an artifact from the build state."""
        con = self.connect_to_database()
        try:
            cur = con.cursor()
            cur.execute(
                """
                delete from artifacts where artifact = ?
            """,
                [artifact_name],
            )
            con.commit()
        finally:
            con.close()

    def _any_sources_are_dirty(self, cur, sources):
        """Given a list of sources this checks if any of them are marked
        as dirty.
        """
        sources = [self.to_source_filename(x) for x in sources]
        if not sources:
            return False

        cur.execute(
            """
            select source from dirty_sources where source in (%s) limit 1
        """
            % ", ".join(["?"] * len(sources)),
            sources,
        )
        return cur.fetchone() is not None

    @staticmethod
    def _get_artifact_config_hash(cur, artifact_name):
        """Returns the artifact's config hash."""
        cur.execute(
            """
            select config_hash from artifact_config_hashes
             where artifact = ?
        """,
            [artifact_name],
        )
        rv = cur.fetchone()
        return rv[0] if rv else None

    def check_artifact_is_current(self, artifact_name, sources, config_hash):
        con = self.connect_to_database()
        cur = con.cursor()
        try:
            # The artifact config changed
            if config_hash != self._get_artifact_config_hash(cur, artifact_name):
                return False

            # If one of our source files is explicitly marked as dirty in the
            # build state, we are not current.
            if self._any_sources_are_dirty(cur, sources):
                return False

            # If we do have an already existing artifact, we need to check if
            # any of the source files we depend on changed.
            for _, info in self._iter_artifact_dependency_infos(
                cur, artifact_name, sources
            ):
                # if we get a missing source info it means that we never
                # saw this before.  This means we need to build it.
                if info is None:
                    return False

                if isinstance(info, VirtualSourceInfo):
                    new_vinfo = self.get_virtual_source_info(info.path)
                    if not info.unchanged(new_vinfo):
                        return False

                # If the file info is different, then it clearly changed.
                elif not info.unchanged(self.get_file_info(info.filename)):
                    return False

            return True
        finally:
            con.close()

    def iter_unreferenced_artifacts(self, all=False):
        """Finds all unreferenced artifacts in the build folder and yields
        them.
        """
        dst = os.path.join(self.builder.destination_path)

        con = self.connect_to_database()
        cur = con.cursor()

        try:
            for dirpath, dirnames, filenames in os.walk(dst):
                dirnames[:] = [
                    x for x in dirnames if not self.env.is_ignored_artifact(x)
                ]
                for filename in filenames:
                    if self.env.is_ignored_artifact(filename):
                        continue
                    full_path = os.path.join(dst, dirpath, filename)
                    artifact_name = self.artifact_name_from_destination_filename(
                        full_path
                    )

                    if all:
                        yield artifact_name
                        continue

                    cur.execute(
                        """
                        select source from artifacts
                         where artifact = ?
                           and is_primary_source""",
                        [artifact_name],
                    )
                    sources = set(x[0] for x in cur.fetchall())

                    # It's a bad artifact if there are no primary sources
                    # or the primary sources do not exist.
                    if not sources or not any(
                        self.get_file_info(x).exists for x in sources
                    ):
                        yield artifact_name
        finally:
            con.close()

    def iter_artifacts(self):
        """Iterates over all artifact and their file infos.."""
        con = self.connect_to_database()
        try:
            cur = con.cursor()
            cur.execute(
                """
                select distinct artifact from artifacts order by artifact
            """
            )
            rows = cur.fetchall()
            con.close()
            for (artifact_name,) in rows:
                path = self.get_destination_filename(artifact_name)
                info = FileInfo(self.builder.env, path)
                if info.exists:
                    yield artifact_name, info
        finally:
            con.close()

    def vacuum(self):
        """Vacuums the build db."""
        con = self.connect_to_database()
        try:
            con.execute("vacuum")
        finally:
            con.close()


def _describe_fs_path_for_checksum(path):
    """Given a file system path this returns a basic description of what
    this is.  This is used for checksum hashing on directories.
    """
    # This is not entirely correct as it does not detect changes for
    # contents from alternatives.  However for the moment it's good
    # enough.
    if os.path.isfile(path):
        return b"\x01"
    if os.path.isfile(os.path.join(path, "contents.lr")):
        return b"\x02"
    if os.path.isdir(path):
        return b"\x03"
    return b"\x00"


class FileInfo:
    """A file info object holds metainformation of a file so that changes
    can be detected easily.
    """

    def __init__(
        self, env, filename, mtime=None, size=None, checksum=None, is_dir=None
    ):
        self.env = env
        self.filename = filename
        if mtime is not None and size is not None and is_dir is not None:
            self._stat = (mtime, size, is_dir)
        else:
            self._stat = None
        self._checksum = checksum

    def _get_stat(self):
        rv = self._stat
        if rv is not None:
            return rv

        try:
            st = os.stat(self.filename)
            mtime = int(st.st_mtime)
            if stat.S_ISDIR(st.st_mode):
                size = len(os.listdir(self.filename))
                is_dir = True
            else:
                size = int(st.st_size)
                is_dir = False
            rv = mtime, size, is_dir
        except OSError:
            rv = 0, -1, False
        self._stat = rv
        return rv

    @property
    def mtime(self):
        """The timestamp of the last modification."""
        return self._get_stat()[0]

    @property
    def size(self):
        """The size of the file in bytes.  If the file is actually a
        dictionary then the size is actually the number of files in it.
        """
        return self._get_stat()[1]

    @property
    def is_dir(self):
        """Is this a directory?"""
        return self._get_stat()[2]

    @property
    def exists(self):
        return self.size >= 0

    @property
    def checksum(self):
        """The checksum of the file or directory."""
        rv = self._checksum
        if rv is not None:
            return rv

        try:
            h = hashlib.sha1()
            if os.path.isdir(self.filename):
                h.update(b"DIR\x00")
                for filename in sorted(os.listdir(self.filename)):
                    if self.env.is_uninteresting_source_name(filename):
                        continue
                    if isinstance(filename, str):
                        filename = filename.encode("utf-8")
                    h.update(filename)
                    h.update(
                        _describe_fs_path_for_checksum(
                            os.path.join(self.filename, filename.decode("utf-8"))
                        )
                    )
                    h.update(b"\x00")
            else:
                with open(self.filename, "rb") as f:
                    while 1:
                        chunk = f.read(16 * 1024)
                        if not chunk:
                            break
                        h.update(chunk)
            checksum = h.hexdigest()
        except (OSError, IOError):
            checksum = "0" * 40
        self._checksum = checksum
        return checksum

    @property
    def filename_and_checksum(self):
        """Like 'filename:checksum'."""
        return "%s:%s" % (self.filename, self.checksum)

    def unchanged(self, other):
        """Given another file info checks if the are similar enough to
        not consider it changed.
        """
        if not isinstance(other, FileInfo):
            raise TypeError("'other' must be a FileInfo, not %r" % other)

        # If mtime and size match, we skip the checksum comparison which
        # might require a file read which we do not want in those cases.
        # (Except if it's a directory, then we won't do that)
        if not self.is_dir and self.mtime == other.mtime and self.size == other.size:
            return True

        return self.checksum == other.checksum


class VirtualSourceInfo:
    def __init__(self, path, mtime=None, checksum=None):
        self.path = path
        self.mtime = mtime
        self.checksum = checksum

    def unchanged(self, other):
        if not isinstance(other, VirtualSourceInfo):
            raise TypeError("'other' must be a VirtualSourceInfo, not %r" % other)

        if self.path != other.path:
            raise ValueError(
                "trying to compare mismatched virtual paths: "
                "%r.unchanged(%r)" % (self, other)
            )

        return (self.mtime, self.checksum) == (other.mtime, other.checksum)

    def __repr__(self):
        return "VirtualSourceInfo(%r, %r, %r)" % (self.path, self.mtime, self.checksum)


artifacts_row = namedtuple(
    "artifacts_row",
    [
        "artifact",
        "source",
        "source_mtime",
        "source_size",
        "source_checksum",
        "is_dir",
        "is_primary_source",
    ],
)


class Artifact:
    """This class represents a build artifact."""

    def __init__(
        self,
        build_state,
        artifact_name,
        dst_filename,
        sources,
        source_obj=None,
        extra=None,
        config_hash=None,
    ):
        self.build_state = build_state
        self.artifact_name = artifact_name
        self.dst_filename = dst_filename
        self.sources = sources
        self.in_update_block = False
        self.updated = False
        self.source_obj = source_obj
        self.extra = extra
        self.config_hash = config_hash

        self._new_artifact_file = None
        self._pending_update_ops = []

    def __repr__(self):
        return "<%s %r>" % (
            self.__class__.__name__,
            self.dst_filename,
        )

    @property
    def is_current(self):
        """Checks if the artifact is current."""
        # If the artifact does not exist, we're not current.
        if not os.path.isfile(self.dst_filename):
            return False

        return self.build_state.check_artifact_is_current(
            self.artifact_name, self.sources, self.config_hash
        )

    def get_dependency_infos(self):
        return self.build_state.get_artifact_dependency_infos(
            self.artifact_name, self.sources
        )

    def ensure_dir(self):
        """Creates the directory if it does not exist yet."""
        dst_dir = os.path.dirname(self.dst_filename)
        try:
            os.makedirs(dst_dir)
        except OSError:
            pass

    def open(self, mode="rb", encoding=None, ensure_dir=None):
        """Opens the artifact for reading or writing.  This is transaction
        safe by writing into a temporary file and by moving it over the
        actual source in commit.
        """
        if ensure_dir is None:
            ensure_dir = "r" not in mode
        if ensure_dir:
            self.ensure_dir()
        if "r" in mode:
            fn = self._new_artifact_file or self.dst_filename
            return open(fn, mode=mode, encoding=encoding)
        if self._new_artifact_file is None:
            fd, tmp_filename = tempfile.mkstemp(
                dir=os.path.dirname(self.dst_filename), prefix=".__trans"
            )
            os.chmod(tmp_filename, 0o644)
            self._new_artifact_file = tmp_filename
            return os.fdopen(fd, mode)
        return open(self._new_artifact_file, mode=mode, encoding=encoding)

    def replace_with_file(self, filename, ensure_dir=True, copy=False):
        """This is similar to open but it will move over a given named
        file.  The file will be deleted by a rollback or renamed by a
        commit.
        """
        if ensure_dir:
            self.ensure_dir()
        if copy:
            with self.open("wb") as df:
                with open(filename, "rb") as sf:
                    shutil.copyfileobj(sf, df)
        else:
            self._new_artifact_file = filename

    def render_template_into(self, template_name, this, **extra):
        """Renders a template into the artifact.  The default behavior is to
        catch the error and render it into the template with a failure marker.
        """
        rv = self.build_state.env.render_template(
            template_name, self.build_state.pad, this=this, **extra
        )
        with self.open("wb") as f:
            f.write(rv.encode("utf-8") + b"\n")

    def _memorize_dependencies(
        self, dependencies=None, virtual_dependencies=None, for_failure=False
    ):
        """This updates the dependencies recorded for the artifact based
        on the direct sources plus the provided dependencies.  This also
        stores the config hash.

        This normally defers the operation until commit but the `for_failure`
        more will immediately commit into a new connection.
        """

        def operation(con):
            primary_sources = set(
                self.build_state.to_source_filename(x) for x in self.sources
            )

            seen = set()
            rows = []
            for source in chain(self.sources, dependencies or ()):
                source = self.build_state.to_source_filename(source)
                if source in seen:
                    continue
                info = self.build_state.get_file_info(source)
                rows.append(
                    artifacts_row(
                        artifact=self.artifact_name,
                        source=source,
                        source_mtime=info.mtime,
                        source_size=info.size,
                        source_checksum=info.checksum,
                        is_dir=info.is_dir,
                        is_primary_source=source in primary_sources,
                    )
                )

                seen.add(source)

            for v_source in virtual_dependencies or ():
                checksum = v_source.get_checksum(self.build_state.path_cache)
                mtime = v_source.get_mtime(self.build_state.path_cache)
                rows.append(
                    artifacts_row(
                        artifact=self.artifact_name,
                        source=v_source.path,
                        source_mtime=mtime,
                        source_size=None,
                        source_checksum=checksum,
                        is_dir=False,
                        is_primary_source=False,
                    )
                )

            reporter.report_dependencies(rows)

            cur = con.cursor()
            if not for_failure:
                cur.execute(
                    "delete from artifacts where artifact = ?", [self.artifact_name]
                )
            if rows:
                cur.executemany(
                    """
                    insert or replace into artifacts (
                        artifact, source, source_mtime, source_size,
                        source_checksum, is_dir, is_primary_source)
                    values (?, ?, ?, ?, ?, ?, ?)
                """,
                    rows,
                )

            if self.config_hash is None:
                cur.execute(
                    """
                    delete from artifact_config_hashes
                     where artifact = ?
                """,
                    [self.artifact_name],
                )
            else:
                cur.execute(
                    """
                    insert or replace into artifact_config_hashes
                           (artifact, config_hash) values (?, ?)
                """,
                    [self.artifact_name, self.config_hash],
                )

            cur.close()

        if for_failure:
            con = self.build_state.connect_to_database()
            try:
                operation(con)
            except:  # noqa
                con.rollback()
                con.close()
                raise
            con.commit()
            con.close()
        else:
            self._auto_deferred_update_operation(operation)

    def clear_dirty_flag(self):
        """Clears the dirty flag for all sources."""

        def operation(con):
            sources = [self.build_state.to_source_filename(x) for x in self.sources]
            cur = con.cursor()
            cur.execute(
                """
                delete from dirty_sources where source in (%s)
            """
                % ", ".join(["?"] * len(sources)),
                list(sources),
            )
            cur.close()
            reporter.report_dirty_flag(False)

        self._auto_deferred_update_operation(operation)

    def set_dirty_flag(self):
        """Given a list of artifacts this will mark all of their sources
        as dirty so that they will be rebuilt next time.
        """

        def operation(con):
            sources = set()
            for source in self.sources:
                sources.add(self.build_state.to_source_filename(source))

            if not sources:
                return

            cur = con.cursor()
            cur.executemany(
                """
                insert or replace into dirty_sources (source) values (?)
            """,
                [(x,) for x in sources],
            )
            cur.close()

            reporter.report_dirty_flag(True)

        self._auto_deferred_update_operation(operation)

    def _auto_deferred_update_operation(self, f):
        """Helper that defers an update operation when inside an update
        block to a later point.  Otherwise it's auto committed.
        """
        if self.in_update_block:
            self._pending_update_ops.append(f)
            return
        con = self.build_state.connect_to_database()
        try:
            f(con)
        except:  # noqa
            con.rollback()
            raise
        con.commit()

    @contextmanager
    def update(self):
        """Opens the artifact for modifications.  At the start the dirty
        flag is cleared out and if the commit goes through without errors it
        stays cleared.  The setting of the dirty flag has to be done by the
        caller however based on the `exc_info` on the context.
        """
        ctx = self.begin_update()
        try:
            yield ctx
        except:  # pylint: disable=bare-except  # noqa
            exc_info = sys.exc_info()
            self.finish_update(ctx, exc_info)
        else:
            self.finish_update(ctx)

    def begin_update(self):
        """Begins an update block."""
        if self.in_update_block:
            raise RuntimeError("Artifact is already open for updates.")
        self.updated = False
        ctx = Context(self)
        ctx.push()
        self.in_update_block = True
        self.clear_dirty_flag()
        return ctx

    def _commit(self):
        con = None
        try:
            for op in self._pending_update_ops:
                if con is None:
                    con = self.build_state.connect_to_database()
                op(con)

            if self._new_artifact_file is not None:
                os.replace(self._new_artifact_file, self.dst_filename)
                self._new_artifact_file = None

            if con is not None:
                con.commit()
                con.close()
                con = None

            self.build_state.updated_artifacts.append(self)
            self.build_state.builder.failure_controller.clear_failure(
                self.artifact_name
            )
        finally:
            if con is not None:
                con.rollback()
                con.close()

    def _rollback(self):
        if self._new_artifact_file is not None:
            try:
                os.remove(self._new_artifact_file)
            except OSError:
                pass
            self._new_artifact_file = None
        self._pending_update_ops = []

    def finish_update(self, ctx, exc_info=None):
        """Finalizes an update block."""
        if not self.in_update_block:
            raise RuntimeError("Artifact is not open for updates.")
        ctx.pop()
        self.in_update_block = False
        self.updated = True

        # If there was no error, we memoize the dependencies like normal
        # and then commit our transaction.
        if exc_info is None:
            self._memorize_dependencies(
                ctx.referenced_dependencies,
                ctx.referenced_virtual_dependencies.values(),
            )
            self._commit()
            return

        # If an error happened we roll back all changes and record the
        # stacktrace in two locations: we record it on the context so
        # that a called can respond to our failure, and we also persist
        # it so that the dev server can render it out later.
        self._rollback()

        # This is a special form of dependency memorization where we do
        # not prune old dependencies and we just append new ones and we
        # use a new database connection that immediately commits.
        self._memorize_dependencies(
            ctx.referenced_dependencies,
            ctx.referenced_virtual_dependencies.values(),
            for_failure=True,
        )

        ctx.exc_info = exc_info
        self.build_state.notify_failure(self, exc_info)


class PathCache:
    def __init__(self, env):
        self.file_info_cache = {}
        self.source_filename_cache = {}
        self.env = env

    def to_source_filename(self, filename):
        """Given a path somewhere below the environment this will return the
        short source filename that is used internally.  Unlike the given
        path, this identifier is also platform independent.
        """
        key = filename
        rv = self.source_filename_cache.get(key)
        if rv is not None:
            return rv
        folder = os.path.abspath(self.env.root_path)
        if isinstance(folder, str) and not isinstance(filename, str):
            filename = filename.decode(fs_enc)
        filename = os.path.normpath(os.path.join(folder, filename))
        if filename.startswith(folder):
            filename = filename[len(folder) :].lstrip(os.path.sep)
            if os.path.altsep:
                filename = filename.lstrip(os.path.altsep)
        else:
            raise ValueError(
                "The given value (%r) is not below the "
                "source folder (%r)" % (filename, self.env.root_path)
            )
        rv = filename.replace(os.path.sep, "/")
        self.source_filename_cache[key] = rv
        return rv

    def get_file_info(self, filename):
        """Returns the file info for a given file.  This will be cached
        on the generator for the lifetime of it.  This means that further
        accesses to this file info will not cause more IO but it might not
        be safe to use the generator after modifications to the original
        files have been performed on the outside.

        Generally this function can be used to acquire the file info for
        any file on the file system but it should onl be used for source
        files or carefully for other things.

        The filename given can be a source filename.
        """
        fn = os.path.join(self.env.root_path, filename)
        rv = self.file_info_cache.get(fn)
        if rv is None:
            self.file_info_cache[fn] = rv = FileInfo(self.env, fn)
        return rv


class Builder:
    def __init__(self, pad, destination_path, buildstate_path=None, extra_flags=None):
        self.extra_flags = process_extra_flags(extra_flags)
        self.pad = pad
        self.destination_path = os.path.abspath(
            os.path.join(pad.db.env.root_path, destination_path)
        )
        if buildstate_path:
            self.meta_path = buildstate_path
        else:
            self.meta_path = os.path.join(self.destination_path, ".lektor")
        self.failure_controller = FailureController(pad, self.destination_path)

        try:
            os.makedirs(self.meta_path)
            if os.listdir(self.destination_path) != [".lektor"]:
                if not click.confirm(
                    click.style(
                        "The build dir %s hasn't been used before, and other "
                        "files or folders already exist there. If you prune "
                        "(which normally follows the build step), "
                        "they will be deleted. Proceed with building?"
                        % self.destination_path,
                        fg="yellow",
                    )
                ):
                    os.rmdir(self.meta_path)
                    raise click.Abort()
        except OSError:
            pass

        con = self.connect_to_database()
        try:
            create_tables(con)
        finally:
            con.close()

    @property
    def env(self):
        """The environment backing this generator."""
        return self.pad.db.env

    @property
    def buildstate_database_filename(self):
        """The filename for the build state database."""
        return os.path.join(self.meta_path, "buildstate")

    def connect_to_database(self):
        con = sqlite3.connect(
            self.buildstate_database_filename,
            isolation_level=None,
            timeout=10,
            check_same_thread=False,
        )
        cur = con.cursor()
        cur.execute("pragma journal_mode=WAL")
        cur.execute("pragma synchronous=NORMAL")
        con.commit()
        cur.close()
        return con

    def touch_site_config(self):
        """Touches the site config which typically will trigger a rebuild."""
        try:
            os.utime(os.path.join(self.env.root_path, "site.ini"), None)
        except OSError:
            pass

    def find_files(self, query, alt=PRIMARY_ALT, lang=None, limit=50, types=None):
        """Returns a list of files that match the query.  This requires that
        the source info is up to date and is primarily used by the admin to
        show files that exist.
        """
        return find_files(self, query, alt, lang, limit, types)

    def new_build_state(self, path_cache=None):
        """Creates a new build state."""
        if path_cache is None:
            path_cache = PathCache(self.env)
        return BuildState(self, path_cache)

    def get_build_program(self, source, build_state):
        """Finds the right build function for the given source file."""
        for cls, builder in chain(
            reversed(self.env.build_programs), reversed(builtin_build_programs)
        ):
            if isinstance(source, cls):
                return builder(source, build_state)
        raise RuntimeError("I do not know how to build %r" % source)

    def build_artifact(self, artifact, build_func):
        """Various parts of the system once they have an artifact and a
        function to build it, will invoke this function.  This ultimately
        is what builds.

        The return value is the ctx that was used to build this thing
        if it was built, or `None` otherwise.
        """
        is_current = artifact.is_current
        with reporter.build_artifact(artifact, build_func, is_current):
            if not is_current:
                with artifact.update() as ctx:
                    # Upon builing anything we record a dependency to the
                    # project file.  This is not ideal but for the moment
                    # it will ensure that if the file changes we will
                    # rebuild.
                    project_file = self.env.project.project_file
                    if project_file:
                        ctx.record_dependency(project_file)
                    build_func(artifact)
                return ctx
        return None

    @staticmethod
    def update_source_info(prog, build_state):
        """Updates a single source info based on a program.  This is done
        automatically as part of a build.
        """
        info = prog.describe_source_record()
        if info is not None:
            build_state.write_source_info(info)

    def prune(self, all=False):
        """This cleans up data left in the build folder that does not
        correspond to known artifacts.
        """
        path_cache = PathCache(self.env)
        with reporter.build(all and "clean" or "prune", self):
            self.env.plugin_controller.emit("before-prune", builder=self, all=all)
            with self.new_build_state(path_cache=path_cache) as build_state:
                for aft in build_state.iter_unreferenced_artifacts(all=all):
                    reporter.report_pruned_artifact(aft)
                    filename = build_state.get_destination_filename(aft)
                    prune_file_and_folder(filename, self.destination_path)
                    build_state.remove_artifact(aft)
                build_state.prune_source_infos()

            if all:
                build_state.vacuum()
            self.env.plugin_controller.emit("after-prune", builder=self, all=all)

    def build(self, source, path_cache=None):
        """Given a source object, builds it."""
        with self.new_build_state(path_cache=path_cache) as build_state:
            with reporter.process_source(source):
                prog = self.get_build_program(source, build_state)
                self.env.plugin_controller.emit(
                    "before-build",
                    builder=self,
                    build_state=build_state,
                    source=source,
                    prog=prog,
                )
                prog.build()
                if build_state.updated_artifacts:
                    self.update_source_info(prog, build_state)
                self.env.plugin_controller.emit(
                    "after-build",
                    builder=self,
                    build_state=build_state,
                    source=source,
                    prog=prog,
                )
                return prog, build_state

    def get_initial_build_queue(self):
        """Returns the initial build queue as deque."""
        return deque(self.pad.get_all_roots())

    def extend_build_queue(self, queue, prog):
        queue.extend(prog.iter_child_sources())
        for func in self.env.custom_generators:
            queue.extend(func(prog.source) or ())

    def build_all(self):
        """Builds the entire tree.  Returns the number of failures."""
        failures = 0
        path_cache = PathCache(self.env)
        # We keep a dummy connection here that does not do anything which
        # helps us with the WAL handling.  See #144
        con = self.connect_to_database()
        try:
            with reporter.build("build", self):
                self.env.plugin_controller.emit("before-build-all", builder=self)
                to_build = self.get_initial_build_queue()
                while to_build:
                    source = to_build.popleft()
                    prog, build_state = self.build(source, path_cache=path_cache)
                    self.extend_build_queue(to_build, prog)
                    failures += len(build_state.failed_artifacts)
                self.env.plugin_controller.emit("after-build-all", builder=self)
                if failures:
                    reporter.report_build_all_failure(failures)
            return failures
        finally:
            con.close()

    def update_all_source_infos(self):
        """Fast way to update all source infos without having to build
        everything.
        """
        # We keep a dummy connection here that does not do anything which
        # helps us with the WAL handling.  See #144
        con = self.connect_to_database()
        try:
            with reporter.build("source info update", self):
                with self.new_build_state() as build_state:
                    to_build = self.get_initial_build_queue()
                    while to_build:
                        source = to_build.popleft()
                        with reporter.process_source(source):
                            prog = self.get_build_program(source, build_state)
                            self.update_source_info(prog, build_state)
                        self.extend_build_queue(to_build, prog)
                build_state.prune_source_infos()
        finally:
            con.close()



from flask import abort
from flask import Blueprint
from flask import g
from flask import redirect
from flask import render_template
from flask import request
from flask import url_for
from werkzeug.wsgi import extract_path_info

from lektor.admin.utils import fs_path_to_url_path
from lektor.constants import PRIMARY_ALT


bp = Blueprint("dash", __name__, url_prefix="/admin")


endpoints = [
    ("/", "index"),
    ("/publish", "publish"),
    ("/<path>/edit", "edit"),
    ("/<path>/delete", "delete"),
    ("/<path>/preview", "preview"),
    ("/<path>/add-child", "add_child"),
    ("/<path>/upload", "add_attachment"),
]


@bp.route("/edit")
def edit_redirect():
    record = None

    # Find out where we wanted to go to.  We need to chop off the leading
    # /admin on this URL as this is where the admin thinks it's placed.
    path = extract_path_info(
        request.url_root.rstrip("/").rsplit("/", 1)[0], request.args.get("path", "/")
    )

    if path is not None:
        record = g.admin_context.pad.resolve_url_path(path, alt_fallback=False)
    if record is None:
        abort(404)
    path = fs_path_to_url_path(record.path.split("@")[0])
    if record.alt != PRIMARY_ALT:
        path += "+" + record.alt
    return redirect(url_for("dash.edit", path=path))


def generic_endpoint(**kwargs):
    """This function is invoked by all dash endpoints."""
    return render_template("dash.html")


for path, endpoint in endpoints:
    bp.add_url_rule(path, endpoint, generic_endpoint)



import os
import threading
import time
import traceback

from werkzeug.serving import run_simple
from werkzeug.serving import WSGIRequestHandler

from lektor.admin import WebAdmin
from lektor.builder import Builder
from lektor.db import Database
from lektor.reporter import CliReporter
from lektor.utils import portable_popen
from lektor.utils import process_extra_flags
from lektor.watcher import Watcher


class SilentWSGIRequestHandler(WSGIRequestHandler):
    def log(self, type, message, *args):
        pass


class BackgroundBuilder(threading.Thread):
    def __init__(self, env, output_path, prune=True, verbosity=0, extra_flags=None):
        threading.Thread.__init__(self)
        self.env = env
        self.output_path = output_path
        self.prune = prune
        self.verbosity = verbosity
        self.last_build = time.time()
        self.extra_flags = extra_flags

    def build(self, update_source_info_first=False):
        try:
            db = Database(self.env)
            builder = Builder(
                db.new_pad(), self.output_path, extra_flags=self.extra_flags
            )
            if update_source_info_first:
                builder.update_all_source_infos()
            builder.build_all()
            if self.prune:
                builder.prune()
        except Exception:
            traceback.print_exc()
        else:
            self.last_build = time.time()

    def run(self):
        with CliReporter(self.env, verbosity=self.verbosity):
            self.build(update_source_info_first=True)
            with Watcher(self.env, self.output_path) as watcher:
                for ts, _, _ in watcher:
                    if self.last_build is None or ts > self.last_build:
                        self.build()


class DevTools:
    """This provides extra helpers for launching tools such as webpack."""

    def __init__(self, env):
        self.watcher = None
        self.env = env

    def start(self):
        if self.watcher is not None:
            return

        # pylint: disable=import-outside-toplevel
        from lektor import admin

        admin = os.path.dirname(admin.__file__)
        portable_popen(["npm", "install", "."], cwd=admin).wait()

        self.watcher = portable_popen(
            [os.path.join(admin, "node_modules/.bin/webpack"), "--watch"],
            cwd=os.path.join(admin, "static"),
        )

    def stop(self):
        if self.watcher is None:
            return
        self.watcher.kill()
        self.watcher.wait()
        self.watcher = None


def browse_to_address(addr):
    # pylint: disable=import-outside-toplevel
    import webbrowser

    def browse():
        time.sleep(1)
        webbrowser.open("http://%s:%s" % addr)

    t = threading.Thread(target=browse)
    t.daemon = True
    t.start()


def run_server(
    bindaddr,
    env,
    output_path,
    prune=True,
    verbosity=0,
    lektor_dev=False,
    ui_lang="en",
    browse=False,
    extra_flags=None,
):
    """This runs a server but also spawns a background process.  It's
    not safe to call this more than once per python process!
    """
    wz_as_main = os.environ.get("WERKZEUG_RUN_MAIN") == "true"
    in_main_process = not lektor_dev or wz_as_main
    extra_flags = process_extra_flags(extra_flags)

    if in_main_process:
        background_builder = BackgroundBuilder(
            env,
            output_path=output_path,
            prune=prune,
            verbosity=verbosity,
            extra_flags=extra_flags,
        )
        background_builder.daemon = True
        background_builder.start()
        env.plugin_controller.emit(
            "server-spawn", bindaddr=bindaddr, extra_flags=extra_flags
        )

    app = WebAdmin(
        env,
        output_path=output_path,
        verbosity=verbosity,
        debug=lektor_dev,
        ui_lang=ui_lang,
        extra_flags=extra_flags,
    )

    dt = None
    if lektor_dev and not wz_as_main:
        dt = DevTools(env)
        dt.start()

    if browse:
        browse_to_address(bindaddr)

    try:
        return run_simple(
            bindaddr[0],
            bindaddr[1],
            app,
            use_debugger=True,
            threaded=True,
            use_reloader=lektor_dev,
            request_handler=WSGIRequestHandler
            if lektor_dev
            else SilentWSGIRequestHandler,
        )
    finally:
        if dt is not None:
            dt.stop()
        if in_main_process:
            env.plugin_controller.emit("server-stop")



import json
import os
from operator import itemgetter

import pytest

from lektor.admin import WebAdmin
from lektor.admin.utils import eventstream


@pytest.fixture
def children_records_data():
    """Returns test values for children records' `id`, `title`, and `pub_date` fields."""
    return (
        {"id": "1", "title": "1 is the first number", "pub_date": "2016-07-11"},
        {
            "id": "2",
            "title": "Must be the Second item in a row",
            "pub_date": "2017-05-03",
        },
        {"id": "3", "title": "Z is the last letter", "pub_date": "2017-05-03"},
        {"id": "4", "title": "Some random string", "pub_date": "2018-05-21"},
    )


@pytest.fixture(scope="function", autouse=True)
def prepare_stub_data(scratch_project, children_records_data):
    """Creates folders, models, test object and its children records."""
    tree = scratch_project.tree
    with open(os.path.join(tree, "models", "mymodel.ini"), "w", encoding="utf-8") as f:
        f.write("[children]\n" "order_by = -pub_date, title\n")
    with open(
        os.path.join(tree, "models", "mychildmodel.ini"), "w", encoding="utf-8"
    ) as f:
        f.write(
            "[fields.title]\n" "type = string\n" "[fields.pub_date]\n" "type = date"
        )
    os.mkdir(os.path.join(tree, "content", "myobj"))
    with open(
        os.path.join(tree, "content", "myobj", "contents.lr"), "w", encoding="utf-8"
    ) as f:
        f.write("_model: mymodel\n" "---\n" "title: My Test Object\n")
    for record in children_records_data:
        os.mkdir(os.path.join(tree, "content", "myobj", record["id"]))
        with open(
            os.path.join(tree, "content", "myobj", record["id"], "contents.lr"),
            "w",
            encoding="utf-8",
        ) as f:
            f.write(
                "_model: mychildmodel\n"
                "---\n"
                "title: %s\n"
                "---\n"
                "pub_date: %s" % (record["title"], record["pub_date"])
            )


def test_children_sorting_via_api(scratch_project, scratch_env, children_records_data):
    webadmin = WebAdmin(scratch_env, output_path=scratch_project.tree)
    data = json.loads(
        webadmin.test_client().get("/admin/api/recordinfo?path=/myobj").data
    )
    children_records_ids_provided_by_api = list(map(itemgetter("id"), data["children"]))

    records_ordered_by_title = sorted(children_records_data, key=itemgetter("title"))
    ordered_records = sorted(
        records_ordered_by_title, key=itemgetter("pub_date"), reverse=True
    )

    assert (
        list(map(itemgetter("id"), ordered_records))
        == children_records_ids_provided_by_api
    )


def test_eventstream_yield_bytes():
    count = 0

    @eventstream
    def testfunc():
        yield "string"
        yield 5

    for data in testfunc().response:  # pylint: disable=no-member
        count += 1
        assert isinstance(data, bytes)
    assert count >= 2



import traceback

from lektor.constants import PRIMARY_ALT
from lektor.environment.expressions import Expression
from lektor.environment.expressions import FormatExpression
from lektor.i18n import get_i18n_block
from lektor.types.base import Type


def _reflow_and_split_labels(labels):
    rv = []
    for lang, string in labels.items():
        for idx, item in enumerate(string.split(",")):
            try:
                d = rv[idx]
            except LookupError:
                d = {}
                rv.append(d)
            d[lang] = item.strip()
    return rv


def _parse_choices(options):
    s = options.get("choices")
    if not s:
        return None

    choices = []
    items = s.split(",")
    user_labels = get_i18n_block(options, "choice_labels")
    implied_labels = []

    for item in items:
        if "=" in item:
            choice, value = item.split("=", 1)
            choice = choice.strip()
            if choice.isdigit():
                choice = int(choice)
            implied_labels.append(value.strip())
            choices.append(choice)
        else:
            choices.append(item.strip())
            implied_labels.append(item.strip())

    if user_labels:
        rv = list(zip(choices, _reflow_and_split_labels(user_labels)))
    else:
        rv = [(key, {"en": label}) for key, label in zip(choices, implied_labels)]

    return rv


class ChoiceSource:
    def __init__(self, env, options):
        source = options.get("source")
        if source is not None:
            self.source = Expression(env, source)
            self.choices = None
            item_key = options.get("item_key") or "{{ this._id }}"
            item_label = options.get("item_label")
        else:
            self.source = None
            self.choices = _parse_choices(options)
            item_key = options.get("item_key") or "{{ this.0 }}"
            item_label = options.get("item_label")
        self.item_key = FormatExpression(env, item_key)
        if item_label is not None:
            item_label = FormatExpression(env, item_label)
        self.item_label = item_label

    @property
    def has_choices(self):
        return self.source is not None or self.choices is not None

    def iter_choices(self, pad, record=None, alt=PRIMARY_ALT):
        values = {}
        if record is not None:
            values["record"] = record
        if self.choices is not None:
            iterable = self.choices
        else:
            try:
                iterable = self.source.evaluate(pad, alt=alt, values=values)
            except Exception:
                traceback.print_exc()
                iterable = ()

        for item in iterable or ():
            key = self.item_key.evaluate(pad, this=item, alt=alt, values=values)

            # If there is a label expression, use it.  Since in that case
            # we only have one language to fill in, we fill it in for the
            # default language
            if self.item_label is not None:
                label = {
                    "en": self.item_label.evaluate(
                        pad, this=item, alt=alt, values=values
                    )
                }

            # Otherwise we create a proper internationalized key out of
            # our target label
            else:
                if isinstance(item, (tuple, list)) and len(item) == 2:
                    label = item[1]
                elif hasattr(item, "get_record_label_i18n"):
                    label = item.get_record_label_i18n()
                else:
                    label = {"en": item["_id"]}

            yield key, label


class MultiType(Type):
    def __init__(self, env, options):
        Type.__init__(self, env, options)
        self.source = ChoiceSource(env, options)

    def get_labels(self, pad, record=None, alt=PRIMARY_ALT):
        return dict(self.source.iter_choices(pad, record, alt))

    def to_json(self, pad, record=None, alt=PRIMARY_ALT):
        rv = Type.to_json(self, pad, record, alt)
        if self.source.has_choices:
            rv["choices"] = list(self.source.iter_choices(pad, record, alt))
        return rv


class SelectType(MultiType):
    widget = "select"

    def value_from_raw(self, raw):
        if raw.value is None:
            return raw.missing_value("Missing select value")
        return raw.value


class CheckboxesType(MultiType):
    widget = "checkboxes"

    def value_from_raw(self, raw):
        rv = [x.strip() for x in (raw.value or "").split(",")]
        if rv == [""]:
            rv = []
        return rv



import sys
from html import unescape
from pathlib import Path

import pytest


@pytest.fixture
def compile_template(scratch_env):
    def compile_template(source, name="tmpl.html"):
        Path(scratch_env.root_path, "templates", name).write_text(
            source, encoding="utf-8"
        )
        return scratch_env.jinja_env.get_template(name)

    return compile_template


def test_jinja2_feature_autoescape(compile_template):
    tmpl = compile_template("{{ value }}", "tmpl.html")
    rendered = tmpl.render(value="<tag>")
    assert unescape(rendered) == "<tag>"
    assert "<" not in rendered


def test_jinja2_feature_with(compile_template):
    tmpl = compile_template("{% with x = 'good' %}{{ x }}{% endwith %}")
    assert tmpl.render() == "good"


def test_jinja2_feature_do(compile_template):
    tmpl = compile_template(
        "{% set x = ['a'] %}{% do x.append('b') %}{{ x|join('-') }}"
    )
    assert tmpl.render() == "a-b"


def test_no_reference_cycle_in_environment(project):
    env = project.make_env(load_plugins=False)
    # reference count should be two: one from our `env` variable, and
    # another from the argument to sys.getrefcount
    assert sys.getrefcount(env) == 2



import os

import flask
import pytest
from werkzeug.exceptions import HTTPException

from lektor.admin.webui import WebAdmin

app = flask.Flask(__name__)


def test_index_html(tmpdir, env):
    webadmin = WebAdmin(env, output_path=str(tmpdir.mkdir("webadmin")))
    info = webadmin.lektor_info

    def resolve(to_resolve):
        with app.test_request_context(to_resolve):
            return info.resolve_artifact(to_resolve)

    # /dir_with_index_html adds slash, then returns contents of index.html.
    with pytest.raises(HTTPException) as exc:
        resolve("/dir_with_index_html")

    assert exc.value.response.status_code == 301
    assert exc.value.response.headers["Location"] == "dir_with_index_html/"

    artifact_folder = "dir_with_index_html"
    artifact_item = "index.html"
    artifact = "{}/{}".format(artifact_folder, artifact_item)
    artifact_path = os.path.join(
        info.output_path,
        artifact_folder,
        artifact_item,
    )
    assert resolve("/dir_with_index_html/") == (artifact, artifact_path)

    # Same for index.htm as for index.html.
    with pytest.raises(HTTPException) as exc:
        resolve("/dir_with_index_htm")

    assert exc.value.response.status_code == 301
    assert exc.value.response.headers["Location"] == "dir_with_index_htm/"

    artifact_folder = "dir_with_index_htm"
    artifact_item = "index.htm"
    artifact = "{}/{}".format(artifact_folder, artifact_item)
    artifact_path = os.path.join(
        info.output_path,
        artifact_folder,
        artifact_item,
    )
    assert resolve("/dir_with_index_htm/") == (artifact, artifact_path)

    # Empty or absent directory has default resolution behavior.
    artifact_path = os.path.join(info.output_path, "empty")
    assert resolve("/empty") == (None, artifact_path)
    artifact_path = os.path.join(info.output_path, "doesnt_exist")
    assert resolve("/doesnt_exist") == (None, artifact_path)



from datetime import date
from datetime import datetime

from babel.dates import get_timezone
from markupsafe import Markup
from pytz import FixedOffset

from lektor.constants import PRIMARY_ALT
from lektor.i18n import get_i18n_block
from lektor.types.base import Type
from lektor.utils import bool_from_string


class SingleInputType(Type):
    widget = "singleline-text"

    def to_json(self, pad, record=None, alt=PRIMARY_ALT):
        rv = Type.to_json(self, pad, record, alt)
        rv["addon_label_i18n"] = get_i18n_block(self.options, "addon_label") or None
        return rv


class StringType(SingleInputType):
    def value_from_raw(self, raw):
        if raw.value is None:
            return raw.missing_value("Missing string")
        try:
            return raw.value.splitlines()[0].strip()
        except IndexError:
            return u""


class StringsType(Type):
    widget = "multiline-text"

    def value_from_raw(self, raw):
        return [x.strip() for x in (raw.value or "").splitlines()]


class TextType(Type):
    widget = "multiline-text"

    def value_from_raw(self, raw):
        if raw.value is None:
            return raw.missing_value("Missing text")
        return raw.value


class HtmlType(Type):
    widget = "multiline-text"

    def value_from_raw(self, raw):
        if raw.value is None:
            return raw.missing_value("Missing HTML")
        return Markup(raw.value)


class IntegerType(SingleInputType):
    widget = "integer"

    def value_from_raw(self, raw):
        if raw.value is None:
            return raw.missing_value("Missing integer value")
        try:
            return int(raw.value.strip())
        except ValueError:
            try:
                return int(float(raw.value.strip()))
            except ValueError:
                return raw.bad_value("Not an integer")


class FloatType(SingleInputType):
    widget = "float"

    def value_from_raw(self, raw):
        if raw.value is None:
            return raw.missing_value("Missing float value")
        try:
            return float(raw.value.strip())
        except ValueError:
            return raw.bad_value("Not an integer")


class BooleanType(Type):
    widget = "checkbox"

    def to_json(self, pad, record=None, alt=PRIMARY_ALT):
        rv = Type.to_json(self, pad, record, alt)
        rv["checkbox_label_i18n"] = get_i18n_block(self.options, "checkbox_label")
        return rv

    def value_from_raw(self, raw):
        if raw.value is None:
            return raw.missing_value("Missing boolean")
        val = bool_from_string(raw.value.strip().lower())
        if val is None:
            return raw.bad_value("Bad boolean value")
        return val


class DateType(SingleInputType):
    widget = "datepicker"

    def value_from_raw(self, raw):
        if raw.value is None:
            return raw.missing_value("Missing date")
        try:
            return date(*map(int, raw.value.split("-")))
        except Exception:
            return raw.bad_value("Bad date format")


class DateTimeType(SingleInputType):
    widget = "singleline-text"

    def value_from_raw(self, raw):
        if raw.value is None:
            return raw.missing_value("Missing datetime")
        try:
            chunks = raw.value.split(" ")
            date_info = [int(bit) for bit in chunks[0].split("-")]
            time_info = [int(bit) for bit in chunks[1].split(":")]
            datetime_info = date_info + time_info
            result = datetime(*datetime_info)

            if len(chunks) > 2:
                try:
                    tz = get_timezone(chunks[-1])
                except LookupError:
                    if len(chunks[-1]) > 5:
                        chunks[-1] = chunks[-1][-5:]
                    delta = int(chunks[-1][1:3]) * 60 + int(chunks[-1][3:])
                    if chunks[-1][0] == "-":
                        delta *= -1
                    tz = FixedOffset(delta)
                return tz.localize(result)

            return result
        except Exception:
            return raw.bad_value("Bad date format")



import pytest


def test_paginated_children(pad):
    page1 = pad.get("/projects", page_num=1)

    assert page1 is not None
    assert page1["_model"] == "projects"
    assert page1["_template"] == "projects.html"

    assert page1.datamodel.pagination_config.per_page == 4

    assert page1.children.count() == 8
    assert page1.page_num == 1
    assert page1.pagination.items.count() == 4

    children = page1.pagination.items.all()
    assert len(children) == 4
    assert [x["name"] for x in children] == [
        u"Coffee",
        u"Bagpipe",
        u"Master",
        u"Oven",
    ]

    assert ("projects", "_primary", "1") in pad.cache.persistent
    assert ("projects", "_primary", "2") not in pad.cache.persistent

    page2 = pad.get("/projects", page_num=2)

    assert page2.children.count() == 8
    assert page2.page_num == 2
    assert page2.pagination.items.count() == 3

    children = page2.pagination.items.all()
    assert len(children) == 3
    assert [x["name"] for x in children] == [
        u"Postage",
        u"Slave",
        u"Wolf",
    ]

    assert ("projects", "_primary", "2") in pad.cache.persistent


def test_unpaginated_children(pad):
    page_all = pad.get("/projects")

    assert page_all

    assert page_all.pagination.items.count() == 7
    assert page_all.page_num is None

    children = page_all.pagination.items.all()
    assert len(children) == 7
    assert [x["name"] for x in children] == [
        u"Coffee",
        u"Bagpipe",
        u"Master",
        u"Oven",
        u"Postage",
        u"Slave",
        u"Wolf",
    ]


def test_pagination_access(pad):
    page = pad.get("/projects", page_num=1)

    assert page.pagination.page == 1
    assert page.pagination.next.pagination.page == 2
    assert page.pagination.for_page(2).pagination.page == 2

    assert page.pagination.for_page(0) is None
    assert page.pagination.for_page(3) is None

    assert pad.get("/projects@1").page_num == 1
    assert pad.get("/projects@2").page_num == 2
    assert pad.get("/projects@3").page_num == 3


def test_pagination_attributes(pad):
    page = pad.get("/projects", page_num=1)
    assert page.pagination.current is page
    assert page.pagination.next is not None
    assert page.pagination.next.page_num == 2
    assert page.pagination.prev is None
    assert page.pagination.pages == 2
    assert page.pagination.prev_num is None
    assert page.pagination.next_num == 2
    assert page.pagination.has_next
    assert not page.pagination.has_prev

    page = page.pagination.next
    assert page.pagination.current is page
    assert page.pagination.next is None
    assert page.pagination.prev.page_num == 1
    assert not page.pagination.has_next
    assert page.pagination.next is None
    assert page.pagination.next_num is None


def test_url_matching_for_pagination(pad):
    page1 = pad.resolve_url_path("/projects/")
    assert page1.page_num == 1

    page2 = pad.resolve_url_path("/projects/page/2/")
    assert page2.page_num == 2

    page1_explicit = pad.resolve_url_path("/projects/page/1/")
    assert page1_explicit is None


def test_parent_access(pad):
    page2 = pad.resolve_url_path("/projects/page/2/")
    assert page2["_path"] == "/projects"
    assert page2.path == "/projects@2"
    assert page2.page_num == 2

    child = page2.pagination.items.first()
    assert child.parent.path == page2["_path"]
    assert child.parent.page_num is None


def test_pagination_url_paths(pad):
    # Even though this is paginated, getting to the non paginated version
    # just looks like going to the first page.  We do this because it
    # is more convenient for usage.
    project = pad.get("/projects")
    assert project.url_path == "/projects/"

    # However first page looks the same
    project = pad.get("/projects", page_num=1)
    assert project.url_path == "/projects/"

    # second page is different
    project = pad.get("/projects", page_num=2)
    assert project.url_path == "/projects/page/2/"


def test_unpaginated_children_other_alt(pad):
    page_all = pad.get("/projects", alt="de")

    assert page_all.pagination.items.count() == 8
    assert page_all.page_num is None

    children = page_all.pagination.items.all()
    assert len(children) == 8
    assert [x["name"] for x in children] == [
        u"Kaffee",
        u"Dudelsack",
        u"Meister",
        u"Ofen",
        u"Porto",
        u"Sklave",
        u"Wolf",
        u"Zaun",
    ]


def test_prev_next(pad):
    # coffee, bagpipe, filtered, master have seq numbers 9, 8, 7, 6.
    # They disagree with alphabetization, to ensure we use the pagination
    # query order, "-seq".
    bagpipe = pad.get("/projects/bagpipe")
    assert bagpipe.get_siblings().prev_page["_id"] == "coffee"

    # Next child "filtered" is skipped by pagination query, skip to "master".
    assert bagpipe.get_siblings().next_page["_id"] == "master"

    # Postage is on the previous page before oven, but prev / next ignore pages.
    oven = pad.get("/projects/oven")
    assert oven.get_siblings().prev_page["_id"] == "master"
    assert oven.get_siblings().next_page["_id"] == "postage"


def test_url_matching_for_alt_pagination(pad):
    page1 = pad.resolve_url_path("/de/projects/")
    assert page1.alt == "de"
    assert page1.page_num == 1

    page2 = pad.resolve_url_path("/de//projects/page/2/")
    assert page2.alt == "de"
    assert page2.page_num == 2

    page1_explicit = pad.resolve_url_path("/de/projects/page/1/")
    assert page1_explicit is None


def test_pagination_items_filter(pad):
    # This tests that items are excluded from the pagination based on a
    # query if needed.
    blog = pad.get("/blog", page_num=1)
    assert (
        blog.datamodel.pagination_config.items
        == "this.children.filter(F._model == 'blog-post')"
    )

    assert blog.children.count() == 3
    assert sorted(x["_id"] for x in blog.children) == ["dummy.xml", "post1", "post2"]

    assert blog.pagination.items.count() == 2
    # Sort order is pub_date descending, so post 2 is first.
    assert blog.pagination.items.first()["_id"] == "post2"

    dummy = blog.children.get("dummy.xml")
    assert dummy is not None
    assert dummy["_model"] == "none"


def test_virtual_path_behavior(pad):
    # Base record
    blog = pad.get("/blog")
    assert blog.path == "/blog"
    assert blog["_path"] == "/blog"
    assert blog.page_num is None
    assert blog.url_path == "/blog/"
    assert blog.record is blog

    # Record for page 1 which is a bit special
    blog_page1 = pad.get("/blog@1")
    assert blog_page1.path == "/blog@1"
    assert blog_page1["_path"] == "/blog"
    assert blog_page1.url_path == "/blog/"
    assert blog_page1.page_num == 1
    assert blog_page1.parent is pad.root
    assert blog_page1.record is blog

    # Record for page 2 which is slightly less special
    blog_page2 = pad.get("/blog@2")
    assert blog_page2.path == "/blog@2"
    assert blog_page2["_path"] == "/blog"
    assert blog_page2.url_path == "/blog/page/2/"
    assert blog_page2.page_num == 2
    assert blog_page2.parent is pad.root
    assert blog_page2.record is blog

    # Make sure URL generation works as you would expect:
    assert blog.url_to("@1", absolute=True) == "/blog/"
    assert blog.url_to("@2", absolute=True) == "/blog/page/2/"
    assert blog.url_to("..", absolute=True) == "/"
    assert blog_page2.url_to("..", absolute=True) == "/"
    assert blog_page2.url_to("@1", absolute=True) == "/blog/"
    assert blog_page2.url_to("@3", absolute=True) == "/blog/page/3/"


@pytest.fixture
def pagination(pad, page, pages, monkeypatch):
    projects = pad.get("/projects", page_num=1)
    pagination = projects.pagination
    # hack in the desired page and total
    monkeypatch.setattr(pagination, "page", page)
    monkeypatch.setattr(pagination, "total", pages)
    monkeypatch.setattr(pagination, "per_page", 1)
    assert pagination.pages == pages
    return pagination


@pytest.mark.parametrize(
    (
        "page",
        "pages",
        "left_edge",
        "left_current",
        "right_current",
        "right_edge",
        "expected",
    ),
    [
        (1, 9, 0, 0, 0, 0, [1, None]),
        (2, 9, 0, 0, 0, 0, [None, 2, None]),
        (8, 9, 0, 0, 0, 0, [None, 8, None]),
        (9, 9, 0, 0, 0, 0, [None, 9]),
        (1, 9, 1, 0, 0, 1, [1, None, 9]),
        (2, 9, 1, 0, 0, 1, [1, 2, None, 9]),
        (3, 9, 1, 0, 0, 1, [1, None, 3, None, 9]),
        (7, 9, 1, 0, 0, 1, [1, None, 7, None, 9]),
        (8, 9, 1, 0, 0, 1, [1, None, 8, 9]),
        (9, 9, 1, 0, 0, 1, [1, None, 9]),
        (1, 9, 0, 1, 1, 0, [1, 2, None]),
        (2, 9, 0, 1, 1, 0, [1, 2, 3, None]),
        (3, 9, 0, 1, 1, 0, [None, 2, 3, 4, None]),
        (7, 9, 0, 1, 1, 0, [None, 6, 7, 8, None]),
        (8, 9, 0, 1, 1, 0, [None, 7, 8, 9]),
        (9, 9, 0, 1, 1, 0, [None, 8, 9]),
        (1, 9, 1, 1, 1, 1, [1, 2, None, 9]),
        (2, 9, 1, 1, 1, 1, [1, 2, 3, None, 9]),
        (3, 9, 1, 1, 1, 1, [1, 2, 3, 4, None, 9]),
        (4, 9, 1, 1, 1, 1, [1, None, 3, 4, 5, None, 9]),
        (6, 9, 1, 1, 1, 1, [1, None, 5, 6, 7, None, 9]),
        (7, 9, 1, 1, 1, 1, [1, None, 6, 7, 8, 9]),
        (8, 9, 1, 1, 1, 1, [1, None, 7, 8, 9]),
        (9, 9, 1, 1, 1, 1, [1, None, 8, 9]),
    ],
)
def test_iter_pages(
    pagination,
    page,
    left_edge,
    left_current,
    right_current,
    right_edge,
    expected,
):
    assert (
        list(pagination.iter_pages(left_edge, left_current, right_current, right_edge))
        == expected
    )



import os
import warnings
from datetime import datetime
from hashlib import md5
from io import BytesIO

import pytest
from markers import imagemagick

from lektor.imagetools import compute_dimensions
from lektor.imagetools import get_image_info
from lektor.imagetools import is_rotated
from lektor.imagetools import make_image_thumbnail
from lektor.imagetools import ThumbnailMode


def almost_equal(a, b, e=0.00001):
    return abs(a - b) < e


@pytest.fixture(scope="function")
def make_svg():
    def _make_svg(size_unit=None, with_declaration=True, w=100, h=100):
        emotional_face = """<svg xmlns="http://www.w3.org/2000/svg" \
height="{h}{unit}" width="{w}{unit}">
    <circle cx="50" cy="50" r="50" fill="lightgrey"/>
    <circle cx="30" cy="30" r="10" fill="black"/>
    <circle cx="60" cy="30" r="10" fill="black"/>
    <path stroke="black" d="M30 70 l30 0" stroke-width="5"/>
 </svg>""".format(
            unit=size_unit or "", w=w, h=h
        ).encode(
            "utf-8"
        )

        if with_declaration:
            xml_declaration = b'<?xml version="1.0" standalone="yes"?>'
            svg = xml_declaration + emotional_face
        else:
            svg = emotional_face

        return BytesIO(svg)

    return _make_svg


def test_exif(pad):
    image = pad.root.attachments.images.get("test.jpg")
    assert image is not None

    assert image.exif

    assert almost_equal(image.exif.altitude, 779.0293)
    assert almost_equal(image.exif.aperture, 2.275)
    assert image.exif.artist is None
    assert image.exif.camera == "Apple iPhone 6"
    assert image.exif.camera_make == "Apple"
    assert image.exif.camera_model == "iPhone 6"
    assert image.exif.copyright is None
    assert image.exif.created_at == datetime(2015, 12, 6, 11, 37, 38)
    assert image.exif.exposure_time == "1/33"
    assert image.exif.f == u"\u0192/2.2"
    assert almost_equal(image.exif.f_num, 2.2)
    assert image.exif.flash_info == "Flash did not fire, compulsory flash mode"
    assert image.exif.focal_length == "4.2mm"
    assert image.exif.focal_length_35mm == "29mm"
    assert image.exif.iso == 160
    assert almost_equal(image.exif.latitude, 46.6338333)
    assert image.exif.lens == "Apple iPhone 6 back camera 4.15mm f/2.2"
    assert image.exif.lens_make == "Apple"
    assert image.exif.lens_model == "iPhone 6 back camera 4.15mm f/2.2"
    assert almost_equal(image.exif.longitude, 13.4048333)
    assert image.exif.location == (image.exif.latitude, image.exif.longitude)
    assert image.exif.shutter_speed == "1/33"

    assert image.exif.documentname == "testName"
    assert image.exif.description == "testDescription"
    assert image.exif.is_rotated

    assert isinstance(image.exif.to_dict(), dict)

    for key, value in image.exif.to_dict().items():
        assert getattr(image.exif, key) == value


def test_image_attributes(pad):
    for img in (
        "test.jpg",  # base image, exif-rotated
        "test-sof-last.jpg",  # same image but with SOF marker last
        "test-progressive.jpg",  # with progressive encoding, rotated in place
    ):
        image = pad.root.attachments.images.get(img)
        assert image is not None

        assert image.width == 384
        assert image.height == 512
        assert image.format == "jpeg"


def test_is_rotated(pad):
    for img, rotated in (
        ("test.jpg", True),
        ("test-sof-last.jpg", True),
        ("test-progressive.jpg", False),
    ):
        image = pad.root.attachments.images.get(img)
        with open(image.attachment_filename, "rb") as f:
            assert is_rotated(f) == rotated


def test_image_info_svg_declaration(make_svg):
    w, h = 100, 100
    svg_with_xml_decl = make_svg(with_declaration=True, h=h, w=w)
    svg_no_xml_decl = make_svg(with_declaration=False, h=h, w=w)
    info_svg_no_xml_decl = get_image_info(svg_no_xml_decl)
    info_svg_with_xml_decl = get_image_info(svg_with_xml_decl)

    expected = "svg", w, h
    assert info_svg_with_xml_decl == expected
    assert info_svg_no_xml_decl == expected


def test_image_info_svg_length(make_svg):
    w, h = 100, 100
    svg_with_unit_px = make_svg(size_unit="px", w=w, h=h)
    svg_no_unit = make_svg(size_unit=None, w=w, h=h)
    info_with_unit_px = get_image_info(svg_with_unit_px)
    info_no_unit = get_image_info(svg_no_unit)

    expected = "svg", 100, 100
    assert info_with_unit_px == expected
    assert info_no_unit == expected


_SIMILAR_THUMBNAILS = {
    # original dimensions = 384 x 512
    "test@192.jpg": (192, 256),
    "test@x256.jpg": (192, 256),
    "test@256x256.jpg": (192, 256),
}
_DIFFERING_THUMBNAILS = {
    "test@300x100_crop.jpg": (300, 100),
    "test@300x100_stretch.jpg": (300, 100),
}
_THUMBNAILS = _SIMILAR_THUMBNAILS.copy()
_THUMBNAILS.update(_DIFFERING_THUMBNAILS)


def test_thumbnail_dimensions_reported(builder):
    builder.build_all()
    with open(
        os.path.join(builder.destination_path, "index.html"), encoding="utf-8"
    ) as f:
        html = f.read()

    for t, (w, h) in _THUMBNAILS.items():
        assert '<img src="%s" width="%s" height="%s">' % (t, w, h) in html


@imagemagick
def test_thumbnail_dimensions_real(builder):
    builder.build_all()
    for t, dimensions in _THUMBNAILS.items():
        image_file = os.path.join(builder.destination_path, t)
        with open(image_file, "rb") as f:
            _format, width, height = get_image_info(f)
            assert (width, height) == dimensions


@imagemagick
def test_thumbnails_similar(builder):
    builder.build_all()
    hashes = []
    for t in _SIMILAR_THUMBNAILS:
        image_file = os.path.join(builder.destination_path, t)
        with open(image_file, "rb") as f:
            hashes.append(md5(f.read()).hexdigest())
    for i in range(1, len(hashes)):
        assert hashes[i] == hashes[0]


@imagemagick
def test_thumbnails_differing(builder):
    builder.build_all()
    hashes = []
    for t in _DIFFERING_THUMBNAILS:
        image_file = os.path.join(builder.destination_path, t)
        with open(image_file, "rb") as f:
            hashes.append(md5(f.read()).hexdigest())
    for i in range(1, len(hashes)):
        assert hashes[i] != hashes[0]


@imagemagick
def test_thumbnail_quality(builder):
    builder.build_all()
    image_file = os.path.join(builder.destination_path, "test@192x256_q20.jpg")
    # See if the image file with said quality suffix exists
    assert os.path.isfile(image_file)

    image_size = os.path.getsize(image_file)
    # And the filesize is less than 9200 bytes
    assert image_size < 9200


# TODO: delete this when the thumbnails backwards-compatibility period ends
@pytest.mark.skip(reason="future behaviour")
def test_large_thumbnail_returns_original(builder):
    builder.build_all()
    with open(
        os.path.join(builder.destination_path, "index.html"), encoding="utf-8"
    ) as f:
        html = f.read()

    assert '<img alt="original" src="./test.jpg" width="384" height="512">' in html


def test_dimensions():
    # landscape
    w, h = 100, 50
    assert compute_dimensions(50, 50, w, h) == (50, 25)
    assert compute_dimensions(100, 20, w, h) == (40, 20)
    assert compute_dimensions(200, 200, w, h) == (200, 100)
    assert compute_dimensions(500, 200, w, h) == (400, 200)
    # test missing dimension
    assert compute_dimensions(50, None, w, h) == (50, 25)
    assert compute_dimensions(None, 20, w, h) == (40, 20)
    assert compute_dimensions(200, None, w, h) == (200, 100)
    assert compute_dimensions(None, 200, w, h) == (400, 200)
    # test that rounding is half-up
    assert compute_dimensions(49, None, w, h) == (49, 25)
    assert compute_dimensions(51, None, w, h) == (51, 26)

    # portrait
    w, h = 50, 100
    assert compute_dimensions(50, 50, w, h) == (25, 50)
    assert compute_dimensions(20, 100, w, h) == (20, 40)
    assert compute_dimensions(200, 200, w, h) == (100, 200)
    assert compute_dimensions(200, 500, w, h) == (200, 400)
    #
    assert compute_dimensions(None, 50, w, h) == (25, 50)
    assert compute_dimensions(20, None, w, h) == (20, 40)
    assert compute_dimensions(None, 200, w, h) == (100, 200)
    assert compute_dimensions(200, None, w, h) == (200, 400)
    #
    assert compute_dimensions(None, 49, w, h) == (25, 49)
    assert compute_dimensions(None, 51, w, h) == (26, 51)


class DummyContext:
    def __init__(self):
        self.sub_artifacts = []

    def sub_artifact(self, **kwargs):
        def decorate(f):
            self.add_sub_artifact(build_func=f, **kwargs)
            return f

        return decorate

    def add_sub_artifact(self, **kwargs):
        self.sub_artifacts.append(kwargs)


class Test_make_image_thumbnail:
    # pylint: disable=no-self-use

    @pytest.fixture
    def ctx(self):
        return DummyContext()

    @pytest.fixture
    def test_jpg(self, project):
        return os.path.join(project.tree, "content/test.jpg")

    @pytest.fixture
    def svg_image(self, make_svg, tmp_path):
        svg_path = tmp_path / "test.svg"
        svg_path.write_bytes(make_svg().getvalue())
        return str(svg_path)

    @pytest.fixture
    def source_url_path(self):
        return "test.jpg"

    def test_no_width_or_height(self, ctx, test_jpg, source_url_path):
        with pytest.raises(ValueError, match=r"Must specify at least one"):
            make_image_thumbnail(ctx, test_jpg, source_url_path)
        assert len(ctx.sub_artifacts) == 0

    def test_warn_fallback_to_fit(self, ctx, test_jpg, source_url_path):
        with pytest.warns(UserWarning, match=r"mode requires both"):
            make_image_thumbnail(
                ctx, test_jpg, source_url_path, width=80, mode=ThumbnailMode.CROP
            )
        assert len(ctx.sub_artifacts) == 1

    def test_unknown_image_format(self, ctx, project, source_url_path):
        bad_img = os.path.join(project.tree, "content/contents.lr")
        with pytest.raises(RuntimeError, match=r"unknown image"):
            make_image_thumbnail(ctx, bad_img, source_url_path, width=80)
        assert len(ctx.sub_artifacts) == 0

    def test_svg(self, ctx, svg_image, source_url_path):
        rv = make_image_thumbnail(ctx, svg_image, source_url_path, 40, 60)
        assert rv.width == 40
        assert rv.height == 60
        assert rv.url_path == source_url_path
        assert len(ctx.sub_artifacts) == 0

    @pytest.mark.parametrize(
        "w, h, mode, upscale, expect",
        [
            (50, None, ThumbnailMode.FIT, None, (50, 67)),
            (None, 128, ThumbnailMode.FIT, None, (96, 128)),
            (50, 50, ThumbnailMode.CROP, None, (50, 50)),
            (50, 50, ThumbnailMode.STRETCH, None, (50, 50)),
            (512, None, ThumbnailMode.FIT, True, (512, 683)),
            (512, 512, ThumbnailMode.CROP, None, (512, 512)),
            (512, 512, ThumbnailMode.STRETCH, None, (512, 512)),
        ],
    )
    def test_scale(
        self, ctx, test_jpg, source_url_path, w, h, mode, upscale, expect, recwarn
    ):
        warnings.simplefilter("error")  # no warnings should be emitted
        rv = make_image_thumbnail(ctx, test_jpg, source_url_path, w, h, mode, upscale)
        assert (rv.width, rv.height) == expect
        assert rv.url_path != source_url_path
        assert len(ctx.sub_artifacts) == 1

    def test_implicit_upscale(self, ctx, test_jpg, source_url_path):
        with pytest.warns(UserWarning, match=r"image is being scaled up"):
            rv = make_image_thumbnail(ctx, test_jpg, source_url_path, 512)
        assert (rv.width, rv.height) == (512, 683)
        assert rv.url_path == "test@512.jpg"
        assert len(ctx.sub_artifacts) == 1

    @pytest.mark.parametrize("mode", iter(ThumbnailMode))
    def test_upscale_false(self, ctx, test_jpg, source_url_path, mode):
        rv = make_image_thumbnail(
            ctx, test_jpg, source_url_path, 4096, 4069, mode, upscale=False
        )
        assert (rv.width, rv.height) == (384, 512)
        assert rv.url_path == source_url_path
        assert len(ctx.sub_artifacts) == 0



import gc
import warnings
import weakref

import pytest

from lektor.publisher import Command


def test_Command_triggers_no_warnings():
    # This excercises the issue where publishing via rsync resulted
    # in ResourceWarnings about unclosed streams.

    with pytest.warns(None) as record:
        # This is essentially how RsyncPublisher runs rsync.
        with Command(["echo"]) as client:
            for _ in client:
                pass

        # The ResourceWarnings regarding unclosed files we are checking for
        # are issued during finalization.  Without this extra effort,
        # finalization wouldn't happen until after the test completes.
        client_is_alive = weakref.ref(client)
        del client
        if client_is_alive():
            gc.collect()

    if client_is_alive():
        warnings.warn(
            "Unable to trigger garbage collection of Command instance, "
            "so unable to check for warnings issued during finalization."
        )

    for warning in record.list:
        print(warning)
    assert len(record) == 0



import fnmatch
import os
import uuid
from functools import update_wrapper

import jinja2
from babel import dates
from jinja2.loaders import split_template_path

from lektor.constants import PRIMARY_ALT
from lektor.context import config_proxy
from lektor.context import get_asset_url
from lektor.context import get_ctx
from lektor.context import get_locale
from lektor.context import site_proxy
from lektor.context import url_to
from lektor.environment.config import Config
from lektor.environment.config import DEFAULT_CONFIG  # noqa - reexport
from lektor.environment.config import ServerInfo  # noqa - reexport
from lektor.environment.config import update_config_from_ini  # noqa - reexport
from lektor.environment.expressions import Expression  # noqa - reexport
from lektor.environment.expressions import FormatExpression  # noqa - reexport
from lektor.markdown import Markdown
from lektor.packages import load_packages
from lektor.pluginsystem import initialize_plugins
from lektor.pluginsystem import PluginController
from lektor.publisher import builtin_publishers
from lektor.utils import format_lat_long
from lektor.utils import tojson_filter


def _pass_locale(func):
    def new_func(*args, **kwargs):
        if kwargs.get("locale", None) is None:
            kwargs["locale"] = get_locale("en_US")
        return func(*args, **kwargs)

    return update_wrapper(new_func, func)


# Special files that should always be ignored.
IGNORED_FILES = ["thumbs.db", "desktop.ini", "Icon\r"]

# These files are important for artifacts and must not be ignored when
# they are built even though they start with dots.
SPECIAL_ARTIFACTS = [".htaccess", ".htpasswd"]

# Default glob pattern of ignored files.
EXCLUDED_ASSETS = ["_*", ".*"]

# Default glob pattern of included files (higher-priority than EXCLUDED_ASSETS).
INCLUDED_ASSETS = []


def any_fnmatch(filename, patterns):
    for pat in patterns:
        if fnmatch.fnmatch(filename, pat):
            return True

    return False


class CustomJinjaEnvironment(jinja2.Environment):
    def _load_template(self, name, globals):
        ctx = get_ctx()

        try:
            rv = jinja2.Environment._load_template(self, name, globals)
            if ctx is not None:
                filename = rv.filename
                ctx.record_dependency(filename)
            return rv
        except jinja2.TemplateSyntaxError as e:
            if ctx is not None:
                ctx.record_dependency(e.filename)
            raise
        except jinja2.TemplateNotFound as e:
            if ctx is not None:
                # If we can't find the template we want to record at what
                # possible locations the template could exist.  This will help
                # out watcher to pick up templates that will appear in the
                # future.  This assumes the loader is a file system loader.
                for template_name in e.templates:
                    pieces = split_template_path(template_name)
                    for base in self.loader.searchpath:
                        ctx.record_dependency(os.path.join(base, *pieces))
            raise


def lookup_from_bag(*args):
    pieces = ".".join(x for x in args if x)
    return site_proxy.databags.lookup(pieces)


class Environment:
    def __init__(self, project, load_plugins=True, extra_flags=None):
        self.project = project
        self.root_path = os.path.abspath(project.tree)

        self.theme_paths = [
            os.path.join(self.root_path, "themes", theme)
            for theme in self.project.themes
        ]

        if not self.theme_paths:
            # load the directories in the themes directory as the themes
            try:
                for fname in os.listdir(os.path.join(self.root_path, "themes")):
                    f = os.path.join(self.root_path, "themes", fname)
                    if os.path.isdir(f):
                        self.theme_paths.append(f)
            except OSError:
                pass

        template_paths = [
            os.path.join(path, "templates")
            for path in [self.root_path] + self.theme_paths
        ]

        self.jinja_env = CustomJinjaEnvironment(
            autoescape=self.select_jinja_autoescape,
            extensions=["jinja2.ext.do"],
            loader=jinja2.FileSystemLoader(template_paths),
        )

        from lektor.db import F, get_alts  # pylint: disable=import-outside-toplevel

        self.jinja_env.filters.update(
            tojson=tojson_filter,
            latformat=lambda x, secs=True: format_lat_long(lat=x, secs=secs),
            longformat=lambda x, secs=True: format_lat_long(long=x, secs=secs),
            latlongformat=lambda x, secs=True: format_lat_long(secs=secs, *x),
            # By default filters need to be side-effect free.  This is not
            # the case for this one, so we need to make it as a dummy
            # context filter so that jinja2 will not inline it.
            url=jinja2.pass_context(lambda ctx, *a, **kw: url_to(*a, **kw)),
            asseturl=jinja2.pass_context(lambda ctx, *a, **kw: get_asset_url(*a, **kw)),
            markdown=jinja2.pass_context(lambda ctx, *a, **kw: Markdown(*a, **kw)),
        )
        self.jinja_env.globals.update(
            F=F,
            url_to=url_to,
            site=site_proxy,
            config=config_proxy,
            bag=lookup_from_bag,
            get_alts=get_alts,
            get_random_id=lambda: uuid.uuid4().hex,
        )
        self.jinja_env.filters.update(
            datetimeformat=_pass_locale(dates.format_datetime),
            dateformat=_pass_locale(dates.format_date),
            timeformat=_pass_locale(dates.format_time),
        )

        # pylint: disable=import-outside-toplevel
        from lektor.types import builtin_types

        self.types = builtin_types.copy()

        self.publishers = builtin_publishers.copy()

        # The plugins that are loaded for this environment.  This is
        # modified by the plugin controller and registry methods on the
        # environment.
        self.plugin_controller = PluginController(self, extra_flags)
        self.plugins = {}
        self.plugin_ids_by_class = {}
        self.build_programs = []
        self.special_file_assets = {}
        self.special_file_suffixes = {}
        self.custom_url_resolvers = []
        self.custom_generators = []
        self.virtual_sources = {}

        if load_plugins:
            self.load_plugins()
        # pylint: disable=import-outside-toplevel
        from lektor.db import siblings_resolver

        self.virtualpathresolver("siblings")(siblings_resolver)

    @property
    def asset_path(self):
        return os.path.join(self.root_path, "assets")

    @property
    def temp_path(self):
        return os.path.join(self.root_path, "temp")

    def load_plugins(self):
        """Loads the plugins."""
        load_packages(self)
        initialize_plugins(self)

    def load_config(self):
        """Loads the current config."""
        return Config(self.project.project_file)

    def new_pad(self):
        """Convenience function to create a database and pad."""
        from lektor.db import Database  # pylint: disable=import-outside-toplevel

        return Database(self).new_pad()

    def is_uninteresting_source_name(self, filename):
        """These files are ignored when sources are built into artifacts."""
        fn = filename.lower()
        if fn in SPECIAL_ARTIFACTS:
            return False

        proj = self.project
        if any_fnmatch(filename, INCLUDED_ASSETS + proj.included_assets):
            # Included by the user's project config, thus not uninteresting.
            return False
        return any_fnmatch(filename, EXCLUDED_ASSETS + proj.excluded_assets)

    @staticmethod
    def is_ignored_artifact(asset_name):
        """This is used by the prune tool to figure out which files in the
        artifact folder should be ignored.
        """
        fn = asset_name.lower()
        if fn in SPECIAL_ARTIFACTS:
            return False
        return fn[:1] == "." or fn in IGNORED_FILES

    def render_template(self, name, pad=None, this=None, values=None, alt=None):
        ctx = self.make_default_tmpl_values(pad, this, values, alt, template=name)
        return self.jinja_env.get_or_select_template(name).render(ctx)

    def make_default_tmpl_values(
        self, pad=None, this=None, values=None, alt=None, template=None
    ):
        values = dict(values or ())

        # If not provided, pick the alt from the provided "this" object.
        # As there is no mandatory format for it, we make sure that we can
        # deal with a bad attribute there.
        if alt is None:
            if this is not None:
                alt = getattr(this, "alt", None)
                if not isinstance(alt, str):
                    alt = None
            if alt is None:
                alt = PRIMARY_ALT

        # This is already a global variable but we can inject it as a
        # local override if available.
        if pad is None:
            ctx = get_ctx()
            if ctx is not None:
                pad = ctx.pad
        if pad is not None:
            values["site"] = pad
        if this is not None:
            values["this"] = this
        if alt is not None:
            values["alt"] = alt
        self.plugin_controller.emit(
            "process-template-context", context=values, template=template
        )
        return values

    @staticmethod
    def select_jinja_autoescape(filename):
        if filename is None:
            return False
        return filename.endswith((".html", ".htm", ".xml", ".xhtml"))

    def resolve_custom_url_path(self, obj, url_path):
        for resolver in self.custom_url_resolvers:
            rv = resolver(obj, url_path)
            if rv is not None:
                return rv
        return None

    # -- methods for the plugin system

    def add_build_program(self, cls, program):
        self.build_programs.append((cls, program))

    def add_asset_type(self, asset_cls, build_program):
        self.build_programs.append((asset_cls, build_program))
        self.special_file_assets[asset_cls.source_extension] = asset_cls
        if asset_cls.artifact_extension:
            cext = asset_cls.source_extension + asset_cls.artifact_extension
            self.special_file_suffixes[cext] = asset_cls.source_extension

    def add_publisher(self, scheme, publisher):
        if scheme in self.publishers:
            raise RuntimeError('Scheme "%s" is already registered.' % scheme)
        self.publishers[scheme] = publisher

    def add_type(self, type):
        name = type.name
        if name in self.types:
            raise RuntimeError('Type "%s" is already registered.' % name)
        self.types[name] = type

    def virtualpathresolver(self, prefix):
        def decorator(func):
            if prefix in self.virtual_sources:
                raise RuntimeError('Prefix "%s" is already registered.' % prefix)
            self.virtual_sources[prefix] = func
            return func

        return decorator

    def urlresolver(self, func):
        self.custom_url_resolvers.append(func)
        return func

    def generator(self, func):
        self.custom_generators.append(func)
        return func
